import argparse
import os
import random
import threading
import time
from contextlib import contextmanager

import numpy as np
import txaio
import yaml
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock
from twisted.internet import reactor

from socs.Lakeshore.Lakeshore372 import LS372


class YieldingLock:
    """A lock protected by a lock.  This braided arrangement guarantees
    that a thread waiting on the lock will get priority over a thread
    that has just released the lock and wants to reacquire it.

    The typical use case is a Process that wants to hold the lock as
    much as possible, but occasionally release the lock (without
    sleeping for long) so another thread can access a resource.  The
    method release_and_acquire() is provided to make this a one-liner.

    """

    def __init__(self, default_timeout=None):
        self.job = None
        self._next = threading.Lock()
        self._active = threading.Lock()
        self._default_timeout = default_timeout

    def acquire(self, timeout=None, job=None):
        if timeout is None:
            timeout = self._default_timeout
        if timeout is None or timeout == 0.:
            kw = {'blocking': False}
        else:
            kw = {'blocking': True, 'timeout': timeout}
        result = False
        if self._next.acquire(**kw):
            if self._active.acquire(**kw):
                self.job = job
                result = True
            self._next.release()
        return result

    def release(self):
        self.job = None
        return self._active.release()

    def release_and_acquire(self, timeout=None):
        job = self.job
        self.release()
        return self.acquire(timeout=timeout, job=job)

    @contextmanager
    def acquire_timeout(self, timeout=None, job='unnamed'):
        result = self.acquire(timeout=timeout, job=job)
        if result:
            try:
                yield result
            finally:
                self.release()
        else:
            yield result


class LS372_Agent:
    """Agent to connect to a single Lakeshore 372 device.

    Args:
        name (ApplicationSession): ApplicationSession for the Agent.
        ip (str): IP Address for the 372 device.
        fake_data (bool, optional): generates random numbers without connecting
            to LS if True.
        dwell_time_delay (int, optional): Amount of time, in seconds, to
            delay data collection after switching channels. Note this time
            should not include the change pause time, which is automatically
            accounted for. Will automatically be reduced to dwell_time - 1
            second if it is set longer than a channel's dwell time. This
            ensures at least one second of data collection at the end of a scan.
        enable_control_chan (bool, optional):
            If True, will read data from the control channel each iteration of
            the acq loop. Defaults to False.
    """

    def __init__(self, agent, name, ip, fake_data=False, dwell_time_delay=0,
                 enable_control_chan=False):

        # self._acq_proc_lock is held for the duration of the acq Process.
        # Tasks that require acq to not be running, at all, should use
        # this lock.
        self._acq_proc_lock = TimeoutLock()

        # self._lock is held by the acq Process only when accessing
        # the hardware but released occasionally so that (short) Tasks
        # may run.  Use a YieldingLock to guarantee that a waiting
        # Task gets activated preferentially, even if the acq thread
        # immediately tries to reacquire.
        self._lock = YieldingLock(default_timeout=5)

        self.name = name
        self.ip = ip
        self.fake_data = fake_data
        self.dwell_time_delay = dwell_time_delay
        self.module = None
        self.thermometers = []

        self.log = agent.log
        self.initialized = False
        self.take_data = False
        self.control_chan_enabled = enable_control_chan

        self.agent = agent
        # Registers temperature feeds
        agg_params = {
            'frame_length': 10 * 60  # [sec]
        }
        self.agent.register_feed('temperatures',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    @ocs_agent.param('_')
    def enable_control_chan(self, session, params=None):
        """enable_control_chan()

        **Task** - Enables readout on the control channel (Channel A).

        """
        self.control_chan_enabled = True
        return True, 'Enabled control channel'

    @ocs_agent.param('_')
    def disable_control_chan(self, session, params=None):
        """disable_control_chan()

        **Task** - Disables readout on the control channel (Channel A).

        """
        self.control_chan_enabled = False
        return True, 'Disabled control channel'

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    @ocs_agent.param('acq_params', type=dict, default=None)
    @ocs_agent.param('force', default=False, type=bool)
    @ocs_agent.param('configfile', type=str, default=None)
    def init_lakeshore(self, session, params=None):
        """init_lakeshore(auto_acquire=False, acq_params=None, force=False, \
                          configfile=None)

        **Task** - Perform first time setup of the Lakeshore 372 communication.

        Parameters:
            auto_acquire (bool, optional): Default is False. Starts data
                acquisition after initialization if True.
            acq_params (dict, optional): Params to pass to acq process if
                auto_acquire is True.
            force (bool, optional): Force initialization, even if already
                initialized. Defaults to False.
            configfile (str, optional): .yaml file for initializing 372 channel
                settings

        """
        if params is None:
            params = {}
        if self.initialized and not params.get('force', False):
            self.log.info("Lakeshore already initialized. Returning...")
            return True, "Already initialized"

        with self._lock.acquire_timeout(job='init') as acquired1, \
                self._acq_proc_lock.acquire_timeout(timeout=0., job='init') \
                as acquired2:
            if not acquired1:
                self.log.warn(f"Could not start init because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"
            if not acquired2:
                self.log.warn(f"Could not start init because "
                              f"{self._acq_proc_lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            if self.fake_data:
                self.res = random.randrange(1, 1000)
                session.add_message("No initialization since faking data")
                self.thermometers = ["thermA", "thermB"]
            else:
                try:
                    self.module = LS372(self.ip)
                except ConnectionError:
                    self.log.error("Could not connect to the LS372. Exiting.")
                    reactor.callFromThread(reactor.stop)
                    return False, 'Lakeshore initialization failed'
                except Exception as e:
                    self.log.error(f"Unhandled exception encountered: {e}")
                    reactor.callFromThread(reactor.stop)
                    return False, 'Lakeshore initialization failed'

                print("Initialized Lakeshore module: {!s}".format(self.module))
                session.add_message("Lakeshore initilized with ID: %s" % self.module.id)

                self.thermometers = [channel.name for channel in self.module.channels]

            self.initialized = True

        if params.get('configfile') is not None:
            self.input_configfile(session, params)
            session.add_message("Lakeshore initial configurations uploaded using: %s" % params['configfile'])

        # Start data acquisition if requested
        if params.get('auto_acquire', False):
            self.agent.start('acq', params.get('acq_params', None))

        return True, 'Lakeshore module initialized.'

    @ocs_agent.param('sample_heater', default=False, type=bool)
    @ocs_agent.param('run_once', default=False, type=bool)
    def acq(self, session, params=None):
        """acq(sample_heater=False, run_once=False)

        **Process** - Acquire data from the Lakeshore 372.

        Parameters:
            sample_heater (bool, optional): Default is False. Will record
                values from the sample heater, typically used to servo a DR if
                True.

        Notes:
            The most recent data collected is stored in session data in the
            structure::

                >>> response.session['data']
                {"fields":
                    {"Channel_05": {"T": 293.644, "R": 33.752, "timestamp": 1601924482.722671},
                     "Channel_06": {"T": 0, "R": 1022.44, "timestamp": 1601924499.5258765},
                     "Channel_08": {"T": 0, "R": 1026.98, "timestamp": 1601924494.8172355},
                     "Channel_01": {"T": 293.41, "R": 108.093, "timestamp": 1601924450.9315426},
                     "Channel_02": {"T": 293.701, "R": 30.7398, "timestamp": 1601924466.6130798},
                     "control": {"T": 293.701, "R": 30.7398, "timestamp": 1601924466.6130798}
                    }
                }

        """
        pm = Pacemaker(10, quantize=True)

        with self._acq_proc_lock.acquire_timeout(timeout=0, job='acq') \
                as acq_acquired, \
                self._lock.acquire_timeout(job='acq') as acquired:
            if not acq_acquired:
                self.log.warn(f"Could not start Process because "
                              f"{self._acq_proc_lock.job} is already running")
                return False, "Could not acquire lock"
            if not acquired:
                self.log.warn(f"Could not start Process because "
                              f"{self._lock.job} is holding the lock")
                return False, "Could not acquire lock"

            session.set_status('running')
            self.log.info("Starting data acquisition for {}".format(self.agent.agent_address))
            previous_channel = None
            last_release = time.time()

            session.data = {"fields": {}}

            self.take_data = True
            while self.take_data:
                pm.sleep()

                # Relinquish sampling lock occasionally.
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self._lock.release_and_acquire(timeout=10):
                        self.log.warn(f"Failed to re-acquire sampling lock, "
                                      f"currently held by {self._lock.job}.")
                        continue

                if self.fake_data:
                    data = {
                        'timestamp': time.time(),
                        'block_name': 'fake-data',
                        'data': {}
                    }
                    for therm in self.thermometers:
                        reading = np.random.normal(self.res, 20)
                        data['data'][therm] = reading
                    time.sleep(.1)

                else:
                    active_channel = self.module.get_active_channel()

                    # The 372 reports the last updated measurement repeatedly
                    # during the "pause change time", this results in several
                    # stale datapoints being recorded. To get around this we
                    # query the pause time and skip data collection during it
                    # if the channel has changed (as it would if autoscan is
                    # enabled.)
                    if previous_channel != active_channel:
                        if previous_channel is not None:
                            pause_time = active_channel.get_pause()
                            self.log.debug("Pause time for {c}: {p}",
                                           c=active_channel.channel_num,
                                           p=pause_time)

                            dwell_time = active_channel.get_dwell()
                            self.log.debug("User set dwell_time_delay: {p}",
                                           p=self.dwell_time_delay)

                            # Check user set dwell time isn't too long
                            if self.dwell_time_delay > dwell_time:
                                self.log.warn("WARNING: User set dwell_time_delay of "
                                              + "{delay} s is larger than channel "
                                              + "dwell time of {chan_time} s. If "
                                              + "you are autoscanning this will "
                                              + "cause no data to be collected. "
                                              + "Reducing dwell time delay to {s} s.",
                                              delay=self.dwell_time_delay,
                                              chan_time=dwell_time,
                                              s=dwell_time - 1)
                                total_time = pause_time + dwell_time - 1
                            else:
                                total_time = pause_time + self.dwell_time_delay

                            for i in range(total_time):
                                self.log.debug("Sleeping for {t} more seconds...",
                                               t=total_time - i)
                                time.sleep(1)

                        # Track the last channel we measured
                        previous_channel = self.module.get_active_channel()

                    current_time = time.time()
                    data = {
                        'timestamp': current_time,
                        'block_name': active_channel.name,
                        'data': {}
                    }

                    # Collect both temperature and resistance values from each Channel
                    channel_str = active_channel.name.replace(' ', '_')
                    temp_reading = self.module.get_temp(unit='kelvin',
                                                        chan=active_channel.channel_num)
                    res_reading = self.module.get_temp(unit='ohms',
                                                       chan=active_channel.channel_num)

                    # For data feed
                    data['data'][channel_str + '_T'] = temp_reading
                    data['data'][channel_str + '_R'] = res_reading
                    session.app.publish_to_feed('temperatures', data)
                    self.log.debug("{data}", data=session.data)

                    # For session.data
                    field_dict = {channel_str: {"T": temp_reading,
                                                "R": res_reading,
                                                "timestamp": current_time}}
                    session.data['fields'].update(field_dict)

                    # Also queries control channel if enabled
                    if self.control_chan_enabled:
                        temp = self.module.get_temp(unit='kelvin', chan=0)
                        res = self.module.get_temp(unit='ohms', chan=0)
                        cur_time = time.time()
                        data = {
                            'timestamp': time.time(),
                            'block_name': 'control_chan',
                            'data': {
                                'control_T': temp,
                                'control_R': res
                            }
                        }
                        session.app.publish_to_feed('temperatures', data)
                        self.log.debug("{data}", data=session.data)
                        # Updates session data w/ control field
                        session.data['fields'].update({
                            'control': {
                                'T': temp, 'R': res, 'timestamp': cur_time
                            }
                        })

                if params.get("sample_heater", False):
                    # Sample Heater
                    heater = self.module.sample_heater
                    hout = heater.get_sample_heater_output()

                    current_time = time.time()
                    htr_data = {
                        'timestamp': current_time,
                        'block_name': "heaters",
                        'data': {}
                    }
                    htr_data['data']['sample_heater_output'] = hout

                    session.app.publish_to_feed('temperatures', htr_data)

                if params['run_once']:
                    break

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        Stops acq process.
        """
        if self.take_data:
            session.set_status('stopping')
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'

    @ocs_agent.param('heater', type=str)
    @ocs_agent.param('range')
    @ocs_agent.param('wait', type=float, default=0)
    def set_heater_range(self, session, params):
        """set_heater_range(heater, range, wait=0)

        **Task** - Adjust the heater range for servoing cryostat. Wait for a
        specified amount of time after the change.

        Parameters:
            heater (str): Name of heater to set range for, either 'sample' or
                'still'.
            range (str, float): see arguments in
                :func:`socs.Lakeshore.Lakeshore372.Heater.set_heater_range`
            wait (float, optional): Amount of time to wait after setting the
                heater range. This allows the servo time to adjust to the new range.

        """
        with self._lock.acquire_timeout(job='set_heater_range') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            heater_string = params.get('heater', 'sample')
            if heater_string.lower() == 'sample':
                heater = self.module.sample_heater
            elif heater_string.lower() == 'still':
                heater = self.module.still_heater

            current_range = heater.get_heater_range()
            self.log.debug(f"Current heater range: {current_range}")

            if params['range'] == current_range:
                print("Current heater range matches commanded value. Proceeding unchanged.")
            else:
                heater.set_heater_range(params['range'])
                time.sleep(params.get('wait', 0))

        return True, f'Set {heater_string} heater range to {params["range"]}'

    @ocs_agent.param('channel', type=int, check=lambda x: 1 <= x <= 16)
    @ocs_agent.param('mode', type=str, choices=['current', 'voltage'])
    def set_excitation_mode(self, session, params):
        """set_excitation_mode(channel, mode)

        **Task** - Set the excitation mode of a specified channel.

        Parameters:
            channel (int): Channel to set the excitation mode for. Valid values
                are 1-16.
            mode (str): Excitation mode. Possible modes are 'current' or
                'voltage'.

        """
        with self._lock.acquire_timeout(job='set_excitation_mode') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            self.module.channels[params['channel']].set_excitation_mode(params['mode'])
            session.add_message(f'post message in agent for Set channel {params["channel"]} excitation mode to {params["mode"]}')
            print(f'print statement in agent for Set channel {params["channel"]} excitation mode to {params["mode"]}')

        return True, f'return text for Set channel {params["channel"]} excitation mode to {params["mode"]}'

    @ocs_agent.param('channel', type=int, check=lambda x: 1 <= x <= 16)
    @ocs_agent.param('value', type=float)
    def set_excitation(self, session, params):
        """set_excitation(channel, value)

        **Task** - Set the excitation voltage/current value of a specified
        channel.

        Parameters:
            channel (int): Channel to set the excitation for. Valid values
                are 1-16.
            value (float): Excitation value in volts or amps depending on set
                excitation mode. See
                :func:`socs.Lakeshore.Lakeshore372.Channel.set_excitation`

        """
        with self._lock.acquire_timeout(job='set_excitation') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            current_excitation = self.module.channels[params['channel']].get_excitation()
            mode = self.module.channels[params["channel"]].get_excitation_mode()
            units = 'amps' if mode == 'current' else 'volts'

            if params['value'] == current_excitation:
                session.add_message(f'Channel {params["channel"]} excitation {mode} already set to {params["value"]} {units}')
            else:
                self.module.channels[params['channel']].set_excitation(params['value'])
                session.add_message(f'Set channel {params["channel"]} excitation {mode} to {params["value"]} {units}')

        return True, f'Set channel {params["channel"]} excitation to {params["value"]} {units}'

    @ocs_agent.param('channel', type=int, check=lambda x: 1 <= x <= 16)
    def get_excitation(self, session, params):
        """get_excitation(channel)

        **Task** - Get the excitation voltage/current value of a specified
        channel.

        Parameters:
            channel (int): Channel to get the excitation for. Valid values
                are 1-16.
        """
        with self._lock.acquire_timeout(job='get_excitation') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            current_excitation = self.module.channels[params["channel"]].get_excitation()
            mode = self.module.channels[params["channel"]].get_excitation_mode()
            units = 'amps' if mode == 'current' else 'volts'
            session.add_message(f'Channel {params["channel"]} excitation {mode} is {current_excitation} {units}')
            session.data = {"excitation": current_excitation}

        return True, f'Channel {params["channel"]} excitation {mode} is {current_excitation} {units}'

    @ocs_agent.param('channel', type=int, check=lambda x: 1 <= x <= 16)
    @ocs_agent.param('resistance_range', type=float)
    def set_resistance_range(self, session, params):
        """set_resistance_range(channel, resistance_range)

        **Task** - Set the resistance range for a specified channel.

        Parameters:
            channel (int): Channel to set the resistance range for. Valid values
                are 1-16.
            resistance_range (float): range in ohms we want to measure. Doesn't
                need to be exactly one of the options on the lakeshore, will select
                closest valid range, though note these are in increments of 2, 6.32,
                20, 63.2, etc.

        Notes:
            If autorange is on when you change the resistance range, it may try to change
            it to another value.
        """
        with self._lock.acquire_timeout(job='set_resistance_range') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            current_resistance_range = self.module.channels[params['channel']].get_resistance_range()

            if params['resistance_range'] == current_resistance_range:
                session.add_message(f'Channel {params["channel"]} resistance_range already set to {params["resistance_range"]}')
            else:
                self.module.channels[params['channel']].set_resistance_range(params['resistance_range'])
                session.add_message(f'Set channel {params["channel"]} resistance range to {params["resistance_range"]}')

        return True, f'Set channel {params["channel"]} resistance range to {params["resistance_range"]}'

    @ocs_agent.param('channel', type=int, check=lambda x: 1 <= x <= 16)
    def get_resistance_range(self, session, params):
        """get_resistance_range(channel)

        **Task** - Get the resistance range for a specified channel.

        Parameters:
            channel (int): Channel to get the resistance range for. Valid values
                are 1-16.
        """
        with self._lock.acquire_timeout(job='get_resistance_range') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            current_resistance_range = self.module.channels[params['channel']].get_resistance_range()
            session.add_message(f'Channel {params["channel"]} resistance range is {current_resistance_range}')
            session.data = {"resistance_range": current_resistance_range}

        return True, f'Channel {params["channel"]} resistance range is {current_resistance_range}'

    @ocs_agent.param('channel', type=int, check=lambda x: 1 <= x <= 16)
    @ocs_agent.param('dwell', type=int, check=lambda x: 1 <= x <= 200)
    def set_dwell(self, session, params):
        """set_dwell(channel, dwell)

        **Task** - Set the autoscanning dwell time for a particular channel.

        Parameters:
            channel (int): Channel to set the dwell time for. Valid values
                are 1-16.
            dwell (int): Dwell time in seconds, type is int and must be in the
                range 1-200 inclusive.
        """
        with self._lock.acquire_timeout(job='set_dwell') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            self.module.channels[params["channel"]].set_dwell(params["dwell"])
            session.add_message(f'Set dwell to {params["dwell"]}')

        return True, f'Set channel {params["channel"]} dwell time to {params["dwell"]}'

    @ocs_agent.param('channel', type=int, check=lambda x: 1 <= x <= 16)
    def get_dwell(self, session, params):
        """get_dwell(channel)

        **Task** - Get the autoscanning dwell time for a particular channel.

        Parameters:
            channel (int): Channel to get the dwell time for. Valid values
                are 1-16.
        """
        with self._lock.acquire_timeout(job='set_dwell') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            current_dwell = self.module.channels[params["channel"]].get_dwell()
            session.add_message(f'Dwell time for channel {params["channel"]} is {current_dwell}')
            session.data = {"dwell_time": current_dwell}

        return True, f'Channel {params["channel"]} dwell time is {current_dwell}'

    @ocs_agent.param('P', type=int)
    @ocs_agent.param('I', type=int)
    @ocs_agent.param('D', type=int)
    def set_pid(self, session, params):
        """set_pid(P, I, D)

        **Task** - Set the PID parameters for servo control of fridge.

        Parameters:
            P (int): Proportional term for PID loop
            I (int): Integral term for the PID loop
            D (int): Derivative term for the PID loop

        Notes:
            Makes a call to :func:`socs.Lakeshore.Lakeshore372.Heater.set_pid`.

        """
        with self._lock.acquire_timeout(job='set_pid') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            self.module.sample_heater.set_pid(params["P"], params["I"], params["D"])
            session.add_message(f'post message text for Set PID to {params["P"]}, {params["I"]}, {params["D"]}')
            print(f'print text for Set PID to {params["P"]}, {params["I"]}, {params["D"]}')

        return True, f'return text for Set PID to {params["P"]}, {params["I"]}, {params["D"]}'

    @ocs_agent.param('channel', type=int)
    def set_active_channel(self, session, params):
        """set_active_channel(channel)

        **Task** - Set the active channel on the LS372.

        Parameters:
            channel (int): Channel to switch readout to. Valid values are 1-16.

        """
        with self._lock.acquire_timeout(job='set_active_channel') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            self.module.set_active_channel(params["channel"])
            session.add_message(f'post message text for set channel to {params["channel"]}')
            print(f'print text for set channel to {params["channel"]}')

        return True, f'return text for set channel to {params["channel"]}'

    @ocs_agent.param('autoscan', type=bool)
    def set_autoscan(self, session, params):
        """set_autoscan(autoscan)

        **Task** - Sets autoscan on the LS372.

        Parameters:
            autoscan (bool): True to enable autoscan, False to disable.

        """
        with self._lock.acquire_timeout(job='set_autoscan') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            if params['autoscan']:
                self.module.enable_autoscan()
                self.log.info('enabled autoscan')
            else:
                self.module.disable_autoscan()
                self.log.info('disabled autoscan')

        return True, 'Set autoscan to {}'.format(params['autoscan'])

    @ocs_agent.param('temperature', type=float, check=lambda x: x < 1)
    def servo_to_temperature(self, session, params):
        """servo_to_temperature(temperature)

        **Task** - Servo to a given temperature using a closed loop PID on a
        fixed channel. This will automatically disable autoscan if enabled.

        Parameters:
            temperature (float): Temperature to servo to in units of Kelvin.

        """
        with self._lock.acquire_timeout(job='servo_to_temperature') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            # Check we're in correct control mode for servo.
            if self.module.sample_heater.mode != 'Closed Loop':
                session.add_message('Changing control to Closed Loop mode for servo.')
                self.module.sample_heater.set_mode("Closed Loop")

            # Check we aren't autoscanning.
            if self.module.get_autoscan() is True:
                session.add_message('Autoscan is enabled, disabling for PID control on dedicated channel.')
                self.module.disable_autoscan()

            # Check we're scanning same channel expected by heater for control.
            if self.module.get_active_channel().channel_num != int(self.module.sample_heater.input):
                session.add_message('Changing active channel to expected heater control input')
                self.module.set_active_channel(int(self.module.sample_heater.input))

            # Check we're setup to take correct units.
            if self.module.get_active_channel().units != 'kelvin':
                session.add_message('Setting preferred units to Kelvin on heater control input.')
                self.module.get_active_channel().set_units('kelvin')

            # Make sure we aren't servoing too high in temperature.
            if params["temperature"] > 1:
                return False, 'Servo temperature is set above 1K. Aborting.'

            self.module.sample_heater.set_setpoint(params["temperature"])

        return True, f'Setpoint now set to {params["temperature"]} K'

    @ocs_agent.param('measurements', type=int)
    @ocs_agent.param('threshold', type=float)
    def check_temperature_stability(self, session, params):
        """check_temperature_stability(measurements, threshold)

        Check servo temperature stability is within threshold.

        Parameters:
            measurements (int): number of measurements to average for stability
                check
            threshold (float): amount within which the average needs to be to
                the setpoint for stability

        """
        with self._lock.acquire_timeout(job='check_temp_stability') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            setpoint = float(self.module.sample_heater.get_setpoint())

            if params is None:
                params = {'measurements': 10, 'threshold': 0.5e-3}

            test_temps = []

            for i in range(params['measurements']):
                test_temps.append(self.module.get_temp())
                time.sleep(.1)  # sampling rate is 10 readings/sec, so wait 0.1 s for a new reading

            mean = np.mean(test_temps)
            session.add_message(f'Average of {params["measurements"]} measurements is {mean} K.')
            print(f'Average of {params["measurements"]} measurements is {mean} K.')

            if np.abs(mean - setpoint) < params['threshold']:
                print("passed threshold")
                session.add_message('Setpoint Difference: ' + str(mean - setpoint))
                session.add_message(f'Average is within {params["threshold"]} K threshold. Proceeding with calibration.')

                return True, f"Servo temperature is stable within {params['threshold']} K"

            else:
                print("we're in the else")
                # adjust_heater(t,rest)

        return False, f"Temperature not stable within {params['threshold']}."

    @ocs_agent.param('heater', type=str, choices=['sample', 'still'])
    @ocs_agent.param('mode', type=str, choices=['Off', 'Monitor Out', 'Open Loop', 'Zone', 'Still', 'Closed Loop', 'Warm up'])
    def set_output_mode(self, session, params=None):
        """set_output_mode(heater, mode)

        **Task** - Set output mode of the heater.

        Parameters:
            heater (str): Name of heater to set range for, either 'sample' or
                'still'.
            mode (str): Specifies mode of heater. Can be "Off", "Monitor Out",
                "Open Loop", "Zone", "Still", "Closed Loop", or "Warm up"

        """
        with self._lock.acquire_timeout(job='set_output_mode') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            if params['heater'].lower() == 'still':
                self.module.still_heater.set_mode(params['mode'])
            if params['heater'].lower() == 'sample':
                self.module.sample_heater.set_mode(params['mode'])
            self.log.info("Set {} output mode to {}".format(params['heater'], params['mode']))

        return True, "Set {} output mode to {}".format(params['heater'], params['mode'])

    @ocs_agent.param('heater', type=str, choices=['sample', 'still'])
    @ocs_agent.param('output', type=float)
    @ocs_agent.param('display', type=str, choices=['current', 'power'], default=None)
    def set_heater_output(self, session, params=None):
        """set_heater_output(heater, output, display=None)

        **Task** - Set display type and output of the heater.

        Parameters:
            heater (str): Name of heater to set range for, either 'sample' or
                'still'.
            output (float): Specifies heater output value. For possible values see
                :func:`socs.Lakeshore.Lakeshore372.Heater.set_heater_output`
            display (str, optional): Specifies heater display type. Can be
                "current" or "power". If None, heater display is not reset
                before setting output.

        Notes:
            For the still heater this sets the still heater manual output, *not*
            the still heater still output. Use
            :func:`LS372_Agent.set_still_output()`
            instead to set the still output.

        """
        with self._lock.acquire_timeout(job='set_heater_output') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            heater = params['heater'].lower()
            output = params['output']

            display = params.get('display', None)

            if heater == 'still':
                self.module.still_heater.set_heater_output(output, display_type=display)
            if heater.lower() == 'sample':
                self.log.info("display: {}\toutput: {}".format(display, output))
                self.module.sample_heater.set_heater_output(output, display_type=display)

            self.log.info("Set {} heater display to {}, output to {}".format(heater, display, output))

            session.set_status('running')

            data = {'timestamp': time.time(),
                    'block_name': '{}_heater_out'.format(heater),
                    'data': {'{}_heater_out'.format(heater): output}
                    }
            session.app.publish_to_feed('temperatures', data)

        return True, "Set {} display to {}, output to {}".format(heater, display, output)

    @ocs_agent.param('output', type=float, check=lambda x: 0 <= x <= 100)
    def set_still_output(self, session, params=None):
        """set_still_output(output)

        **Task** - Set the still output on the still heater. This is different
        than the manual output on the still heater. Use
        :func:`LS372_Agent.set_heater_output()` for that.

        Parameters:
            output (float): Specifies still heater output value as a percentage. Can be any
                number between 0 and 100.

        """
        with self._lock.acquire_timeout(job='set_still_output') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            output = params['output']

            self.module.still_heater.set_still_output(output)

            self.log.info("Set still output to {}".format(output))

            session.set_status('running')

            data = {'timestamp': time.time(),
                    'block_name': 'still_heater_still_out',
                    'data': {'still_heater_still_out': output}
                    }
            session.app.publish_to_feed('temperatures', data)

        return True, "Set still output to {}".format(output)

    @ocs_agent.param('_')
    def get_still_output(self, session, params=None):
        """get_still_output()

        **Task** - Gets the current still output on the still heater.

        Notes:
            The still heater output is stored in the session data
            object in the format::

              >>> response.session['data']
              {"still_heater_still_out": 9.628}

        """
        with self._lock.acquire_timeout(job='get_still_output') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            still_output = self.module.still_heater.get_still_output()

            self.log.info("Current still output is {}".format(still_output))

            session.set_status('running')
            session.data = {"still_heater_still_out": still_output}

        return True, "Current still output is {}".format(still_output)

    @ocs_agent.param('configfile', type=str)
    def input_configfile(self, session, params=None):
        """input_configfile(configfile)

        **Task** - Upload 372 configuration file to initialize channel/device
        settings.

        Parameters:
            configfile (str): name of .yaml config file

        """
        with self._lock.acquire_timeout(job='input_configfile') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            # path to configfile in docker container
            configpath = os.environ.get("OCS_CONFIG_DIR", "/config/")
            configfile = params['configfile']

            ls372configs = os.path.join(configpath, configfile)
            with open(ls372configs) as f:
                config = yaml.safe_load(f)

            ls = self.module
            ls_serial = ls.id.split(',')[2]

            device_config = config[ls_serial]['device_settings']
            ls_chann_settings = config[ls_serial]['channel']

            session.set_status('running')

            # enable/disable autoscan
            if device_config['autoscan'] == 'on':
                ls.enable_autoscan()
                self.log.info("autoscan enabled")
            elif device_config['autoscan'] == 'off':
                ls.disable_autoscan()
                self.log.info("autoscan disabled")

            for i in ls_chann_settings:
                # enable/disable channel
                if ls_chann_settings[i]['enable'] == 'on':
                    ls.channels[i].enable_channel()
                    self.log.info("CH.{channel} enabled".format(channel=i))
                elif ls_chann_settings[i]['enable'] == 'off':
                    ls.channels[i].disable_channel()
                    self.log.info("CH.{channel} disabled".format(channel=i))

                # autorange
                if ls_chann_settings[i]['autorange'] == 'on':
                    ls.channels[i].enable_autorange()
                    self.log.info("autorange on")
                elif ls_chann_settings[i]['autorange'] == 'off':
                    ls.channels[i].disable_autorange()
                    self.log.info("autorange off")

                excitation_mode = ls_chann_settings[i]['excitation_mode']
                ls.channels[i].set_excitation_mode(excitation_mode)
                self.log.info("excitation mode for CH.{channel} set to {exc_mode}".format(channel=i, exc_mode=excitation_mode))

                excitation_value = ls_chann_settings[i]['excitation_value']
                ls.channels[i].set_excitation(excitation_value)
                self.log.info("excitation for CH.{channel} set to {exc}".format(channel=i, exc=excitation_value))

                dwell = ls_chann_settings[i]['dwell']
                ls.channels[i].set_dwell(dwell)
                self.log.info("dwell for CH.{channel} is set to {dwell}".format(channel=i, dwell=dwell))

                pause = ls_chann_settings[i]['pause']
                ls.channels[i].set_pause(pause)
                self.log.info("pause for CH.{channel} is set to {pause}".format(channel=i, pause=pause))

                calibration_curvenum = ls_chann_settings[i]['calibration_curve_num']
                ls.channels[i].set_calibration_curve(calibration_curvenum)
                self.log.info("calibration curve for CH.{channel} set to {cal_curve}".format(channel=i, cal_curve=calibration_curvenum))
                tempco = ls_chann_settings[i]['temperature_coeff']
                ls.channels[i].set_temperature_coefficient(tempco)
                self.log.info("temperature coeff. for CH.{channel} set to {tempco}".format(channel=i, tempco=tempco))

        return True, "Uploaded {}".format(configfile)


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address')
    pgroup.add_argument('--serial-number')
    pgroup.add_argument('--fake-data', type=int, default=0,
                        help='Set non-zero to fake data, without hardware.')
    pgroup.add_argument('--dwell-time-delay', type=int, default=0,
                        help="Amount of time, in seconds, to delay data\
                              collection after switching channels. Note this\
                              time should not include the change pause time,\
                              which is automatically accounted for.\
                              Will automatically be reduced to dwell_time - 1\
                              second if it is set longer than a channel's dwell\
                              time. This ensures at least one second of data\
                              collection at the end of a scan.")
    pgroup.add_argument('--mode', type=str, default='acq',
                        choices=['idle', 'init', 'acq'],
                        help="Starting action for the Agent.")
    pgroup.add_argument('--sample-heater', type=bool, default=False,
                        help='Record sample heater output during acquisition.')
    pgroup.add_argument('--enable-control-chan', action='store_true',
                        help='Enable reading of the control input each acq cycle')
    pgroup.add_argument('--configfile', type=str, help='Yaml file for initializing 372 settings')

    return parser


def main(args=None):
    # For logging
    txaio.use_twisted()
    txaio.make_logger()

    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='Lakeshore372Agent',
                                  parser=parser,
                                  args=args)

    # Automatically acquire data if requested (default)
    init_params = False
    if args.mode == 'init':
        init_params = {'auto_acquire': False,
                       'acq_params': {'sample_heater': args.sample_heater},
                       'configfile': args.configfile}
    elif args.mode == 'acq':
        init_params = {'auto_acquire': True,
                       'acq_params': {'sample_heater': args.sample_heater}}

    # Interpret options in the context of site_config.
    print('I am in charge of device with serial number: %s' % args.serial_number)

    agent, runner = ocs_agent.init_site_agent(args)

    lake_agent = LS372_Agent(agent, args.serial_number, args.ip_address,
                             fake_data=args.fake_data,
                             dwell_time_delay=args.dwell_time_delay,
                             enable_control_chan=args.enable_control_chan)

    agent.register_task('init_lakeshore', lake_agent.init_lakeshore,
                        startup=init_params)
    agent.register_task('set_heater_range', lake_agent.set_heater_range)
    agent.register_task('set_excitation_mode', lake_agent.set_excitation_mode)
    agent.register_task('set_excitation', lake_agent.set_excitation)
    agent.register_task('get_excitation', lake_agent.get_excitation)
    agent.register_task('set_resistance_range', lake_agent.set_resistance_range)
    agent.register_task('get_resistance_range', lake_agent.get_resistance_range)
    agent.register_task('set_dwell', lake_agent.set_dwell)
    agent.register_task('get_dwell', lake_agent.get_dwell)
    agent.register_task('set_pid', lake_agent.set_pid)
    agent.register_task('set_autoscan', lake_agent.set_autoscan)
    agent.register_task('set_active_channel', lake_agent.set_active_channel)
    agent.register_task('servo_to_temperature', lake_agent.servo_to_temperature)
    agent.register_task('check_temperature_stability', lake_agent.check_temperature_stability)
    agent.register_task('set_output_mode', lake_agent.set_output_mode)
    agent.register_task('set_heater_output', lake_agent.set_heater_output)
    agent.register_task('set_still_output', lake_agent.set_still_output)
    agent.register_task('get_still_output', lake_agent.get_still_output)
    agent.register_process('acq', lake_agent.acq, lake_agent._stop_acq)
    agent.register_task('enable_control_chan', lake_agent.enable_control_chan)
    agent.register_task('disable_control_chan', lake_agent.disable_control_chan)
    agent.register_task('input_configfile', lake_agent.input_configfile)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



'''
Copyright © 2018-2019 Pico Technology Ltd.

Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
'''

import ctypes
import time

import numpy as np
from picosdk.functions import adc2mV, assert_pico_ok, splitMSODataFast
from picosdk.ps3000a import ps3000a as ps


class ps3000a():
    def __init__(self, sizeofbuffer, samplerate):
        self.status = {}
        self.chandle = ctypes.c_int16()
        self.interval = int(1. / samplerate * 1e9)  # unit ns

        # Size of capture
        self.sizeOfOneBuffer = sizeofbuffer
        self.numBuffersToCapture = 1
        self.totalSamples = self.sizeOfOneBuffer * self.numBuffersToCapture

        print('sample points: {}, sampling rate: {} (MHz), length: {} (sec)'.format(self.totalSamples, samplerate * 1e-6, self.totalSamples * self.interval * 1e-9))

        # Opens the device/s
        self.status["openunit"] = ps.ps3000aOpenUnit(ctypes.byref(self.chandle), None)

        self.info = {}
        self.info['sample_points'] = self.totalSamples
        self.info['sampling_rate_Hz'] = samplerate
        self.info['length_sec'] = self.totalSamples * self.interval * 1e-9

        try:
            assert_pico_ok(self.status["openunit"])
        except BaseException:

            # powerstate becomes the status number of openunit
            powerstate = self.status["openunit"]

            # If powerstate is the same as 282 then it will run this if statement
            if powerstate == 282:
                # Changes the power input to "PICO_POWER_SUPPLY_NOT_CONNECTED"
                self.status["ChangePowerSource"] = ps.ps3000aChangePowerSource(self.chandle, 282)
            # If the powerstate is the same as 286 then it will run this if statement
            elif powerstate == 286:
                # Changes the power input to "PICO_USB3_0_DEVICE_NON_USB3_0_PORT"
                self.status["ChangePowerSource"] = ps.ps3000aChangePowerSource(self.chandle, 286)
            else:
                raise
            assert_pico_ok(self.status["ChangePowerSource"])

    def close(self):
        # Closes the unit
        # Handle = chandle
        self.status["close"] = ps.ps3000aCloseUnit(self.chandle)
        assert_pico_ok(self.status["close"])

        '''
        ##### Signal Generator
        ### wavetype
        PS3000A_SINE sine wave
        PS3000A_SQUARE square wave
        PS3000A_TRIANGLE triangle wave
        PS3000A_DC_VOLTAGE DC voltage
        The following waveTypes apply to B and MSO models only.
        PS3000A_RAMP_UP rising sawtooth
        PS3000A_RAMP_DOWN falling sawtooth
        PS3000A_SINC sin (x)/x
        PS3000A_GAUSSIAN Gaussian
        PS3000A_HALF_SINE

        ###sweep type
        PS3000A_UP
        PS3000A_DOWN
        PS3000A_UPDOWN
        PS3000A_DOWNUP

        ###
        increment: the amount of frequency increase or decrease in sweep mode
        dwell time: the time for which the sweep stays at each frequency, in seconds
        '''

    def SigGenSingle(self, frequency, ptp=2000000):
        wavetype = ctypes.c_int16(0)
        sweepType = ctypes.c_int32(0)
        triggertype = ctypes.c_int32(0)
        triggerSource = ctypes.c_int32(0)

        offsetVoltage = 0
        pkToPk = ptp  # (uV)
        print('Drive frequency: {} (Hz), PKtoPK: {} (V)'.format(frequency, pkToPk * 1e-6))
        self.info['Drive_frequency_Hz'] = frequency
        self.info['PKtoPk'] = pkToPk * 1e-6

        start_frequency = frequency  # Hz
        stop_frequency = frequency  # Hz
        increment = 0
        dwelltime = 1

        self.status["SetSigGenBuiltIn"] = ps.ps3000aSetSigGenBuiltIn(
            self.chandle, offsetVoltage, pkToPk,
            wavetype, start_frequency, stop_frequency, increment, dwelltime, sweepType,
            0,  # operation
            0,  # shots
            0,  # sweeps
            triggertype,
            triggerSource,
            1,  # extInThreshold
        )
        assert_pico_ok(self.status["SetSigGenBuiltIn"])

    def SigGenSweep(self, start_frequency=100e3, stop_frequency=300e3, increment=5e2, dwelltime=1e-3):
        wavetype = ctypes.c_int16(0)
        sweepType = ctypes.c_int32(0)
        triggertype = ctypes.c_int32(0)
        triggerSource = ctypes.c_int32(0)

        offsetVoltage = 0
        pkToPk = 2000000  # (uV)

        self.status["SetSigGenBuiltIn"] = ps.ps3000aSetSigGenBuiltIn(
            self.chandle, offsetVoltage, pkToPk,
            wavetype, start_frequency, stop_frequency, increment, dwelltime, sweepType,
            0,  # operation
            0,  # shots
            0,  # sweeps
            triggertype,
            triggerSource,
            1,  # extInThreshold
        )
        assert_pico_ok(self.status["SetSigGenBuiltIn"])

    def SetScope(self):
        enabled = 1
        disabled = 0  # noqa: F841
        analogue_offset = 0.0

        # Set up channel A
        # handle = chandle
        # channel = PS3000A_CHANNEL_A = 0
        # enabled = 1
        # coupling type = PS3000A_DC = 1
        # range = PS3000A_2V = 7
        # analogue offset = 0 V
        self.channel_range = ps.PS3000A_RANGE['PS3000A_2V']

        self.status["setChA"] = ps.ps3000aSetChannel(
            self.chandle,
            ps.PS3000A_CHANNEL['PS3000A_CHANNEL_A'],
            enabled,
            ps.PS3000A_COUPLING['PS3000A_DC'],
            self.channel_range,
            analogue_offset
        )
        assert_pico_ok(self.status["setChA"])

    def SetScopeAll(self):
        enabled = 1
        disabled = 0  # noqa: F841
        analogue_offset = 0.0

        # Set up channel A
        # handle = chandle
        # channel = PS3000A_CHANNEL_A = 0
        # enabled = 1
        # coupling type = PS3000A_DC = 1
        # range = PS3000A_2V = 7
        # analogue offset = 0 V
        self.channel_range = ps.PS3000A_RANGE['PS3000A_2V']

        for ch in ['A', 'B', 'C', 'D']:
            self.status['setCh' + ch] = ps.ps3000aSetChannel(
                self.chandle,
                ps.PS3000A_CHANNEL['PS3000A_CHANNEL_' + ch],
                enabled,
                ps.PS3000A_COUPLING['PS3000A_DC'],
                self.channel_range,
                analogue_offset
            )
            assert_pico_ok(self.status['setCh' + ch])

    def SetBuffer(self):
        # Create buffers ready for assigning pointers for data collection
        self.bufferAMax = np.zeros(shape=self.sizeOfOneBuffer, dtype=np.int16)
        self.bufferBMax = np.zeros(shape=self.sizeOfOneBuffer, dtype=np.int16)

        memory_segment = 0

        # Set data buffer location for data collection from channel A
        # handle = chandle
        # source = PS3000A_CHANNEL_A = 0
        # pointer to buffer max = ctypes.byref(bufferAMax)
        # pointer to buffer min = ctypes.byref(bufferAMin)
        # buffer length = maxSamples
        # segment index = 0
        # ratio mode = PS3000A_RATIO_MODE_NONE = 0
        self.status["setDataBuffersA"] = ps.ps3000aSetDataBuffers(
            self.chandle,
            ps.PS3000A_CHANNEL['PS3000A_CHANNEL_A'],
            self.bufferAMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
            None,
            self.sizeOfOneBuffer,
            memory_segment,
            ps.PS3000A_RATIO_MODE['PS3000A_RATIO_MODE_NONE']
        )
        assert_pico_ok(self.status["setDataBuffersA"])

    def SetBufferAll(self):
        # Create buffers ready for assigning pointers for data collection
        self.bufferAMax = np.zeros(shape=self.sizeOfOneBuffer, dtype=np.int16)
        self.bufferBMax = np.zeros(shape=self.sizeOfOneBuffer, dtype=np.int16)
        self.bufferCMax = np.zeros(shape=self.sizeOfOneBuffer, dtype=np.int16)
        self.bufferDMax = np.zeros(shape=self.sizeOfOneBuffer, dtype=np.int16)

        memory_segment = 0

        # Set data buffer location for data collection from channel A
        # handle = chandle
        # source = PS3000A_CHANNEL_A = 0
        # pointer to buffer max = ctypes.byref(bufferAMax)
        # pointer to buffer min = ctypes.byref(bufferAMin)
        # buffer length = maxSamples
        # segment index = 0
        # ratio mode = PS3000A_RATIO_MODE_NONE = 0

        self.status["setDataBuffersA"] = ps.ps3000aSetDataBuffers(
            self.chandle,
            ps.PS3000A_CHANNEL['PS3000A_CHANNEL_A'],
            self.bufferAMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
            None,
            self.sizeOfOneBuffer,
            memory_segment,
            ps.PS3000A_RATIO_MODE['PS3000A_RATIO_MODE_NONE']
        )
        assert_pico_ok(self.status["setDataBuffersA"])

        self.status["setDataBuffersB"] = ps.ps3000aSetDataBuffers(
            self.chandle,
            ps.PS3000A_CHANNEL['PS3000A_CHANNEL_B'],
            self.bufferBMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
            None,
            self.sizeOfOneBuffer,
            memory_segment,
            ps.PS3000A_RATIO_MODE['PS3000A_RATIO_MODE_NONE']
        )
        assert_pico_ok(self.status["setDataBuffersB"])

        self.status["setDataBuffersC"] = ps.ps3000aSetDataBuffers(
            self.chandle,
            ps.PS3000A_CHANNEL['PS3000A_CHANNEL_C'],
            self.bufferCMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
            None,
            self.sizeOfOneBuffer,
            memory_segment,
            ps.PS3000A_RATIO_MODE['PS3000A_RATIO_MODE_NONE']
        )
        assert_pico_ok(self.status["setDataBuffersC"])

        self.status["setDataBuffersD"] = ps.ps3000aSetDataBuffers(
            self.chandle,
            ps.PS3000A_CHANNEL['PS3000A_CHANNEL_D'],
            self.bufferDMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
            None,
            self.sizeOfOneBuffer,
            memory_segment,
            ps.PS3000A_RATIO_MODE['PS3000A_RATIO_MODE_NONE']
        )
        assert_pico_ok(self.status["setDataBuffersD"])

    def Stream(self):
        # Begin streaming mode:
        sampleInterval = ctypes.c_int32(500)
        # print(sampleInterval.value)
        sampleUnits = ps.PS3000A_TIME_UNITS['PS3000A_NS']  # PS3000A_US, PS3000A_NS
        # We are not triggering:
        maxPreTriggerSamples = 0
        autoStopOn = 1
        # No downsampling:
        downsampleRatio = 1
        self.status["runStreaming"] = ps.ps3000aRunStreaming(
            self.chandle,
            ctypes.byref(sampleInterval),
            sampleUnits,
            maxPreTriggerSamples,
            self.totalSamples,
            autoStopOn,
            downsampleRatio,
            ps.PS3000A_RATIO_MODE['PS3000A_RATIO_MODE_NONE'],
            self.sizeOfOneBuffer
        )
        assert_pico_ok(self.status["runStreaming"])

        actualSampleInterval = sampleInterval.value
        # print(sampleInterval.value)
        self.actualSampleIntervalNs = actualSampleInterval * 1000

        print("Capturing at sample interval %s ns" % self.actualSampleIntervalNs)

        # We need a big buffer, not registered with the driver, to keep our complete capture in.
        self.bufferCompleteA = np.zeros(shape=self.totalSamples, dtype=np.int16)
        self.bufferCompleteB = np.zeros(shape=self.totalSamples, dtype=np.int16)
        self.bufferCompleteC = np.zeros(shape=self.totalSamples, dtype=np.int16)
        self.bufferCompleteD = np.zeros(shape=self.totalSamples, dtype=np.int16)
        self.nextSample = 0
        self.autoStopOuter = False
        self.wasCalledBack = False
        # Convert the python function into a C function pointer.
        cFuncPtr = ps.StreamingReadyType(self._streaming_callback)

        # Fetch data from the driver in a loop,
        # copying it out of the registered buffers and into our complete one.

        while self.nextSample < self.totalSamples and not self.autoStopOuter:
            self.wasCalledBack = False
            self.status["getStreamingLastestValues"] = ps.ps3000aGetStreamingLatestValues(
                self.chandle,
                cFuncPtr,
                None
            )
            if not self.wasCalledBack:
                # If we weren't called back by the driver, this means no data is ready.
                # Sleep for a short while before trying
                # again.
                time.sleep(0.01)

        print("Done grabbing values.")
        print("Capturing interval %s sec" % (actualSampleInterval * 1000 * self.totalSamples / 1e12))

    def _streaming_callback(self, handle, noOfSamples, startIndex, overflow, triggerAt, triggered, autoStop, param):
        # global nextSample, autoStopOuter, wasCalledBack
        self.wasCalledBack = True
        destEnd = self.nextSample + noOfSamples
        sourceEnd = startIndex + noOfSamples
        self.bufferCompleteA[self.nextSample:destEnd] = self.bufferAMax[startIndex:sourceEnd]
        self.bufferCompleteB[self.nextSample:destEnd] = self.bufferBMax[startIndex:sourceEnd]
        self.bufferCompleteC[self.nextSample:destEnd] = self.bufferCMax[startIndex:sourceEnd]
        self.bufferCompleteD[self.nextSample:destEnd] = self.bufferDMax[startIndex:sourceEnd]
        self.nextSample += noOfSamples
        if autoStop:
            self.autoStopOuter = True

    def get_value(self):
        # Find maximum ADC count value
        # handle = chandle
        # pointer to value = ctypes.byref(maxADC)
        maxADC = ctypes.c_int16()
        self.status["maximumValue"] = ps.ps3000aMaximumValue(self.chandle, ctypes.byref(maxADC))
        assert_pico_ok(self.status["maximumValue"])

        # Convert ADC counts data to mV
        self.adc2mVChAMax = adc2mV(self.bufferCompleteA, self.channel_range, maxADC)
        self.adc2mVChBMax = adc2mV(self.bufferCompleteB, self.channel_range, maxADC)
        self.adc2mVChCMax = adc2mV(self.bufferCompleteC, self.channel_range, maxADC)
        self.adc2mVChDMax = adc2mV(self.bufferCompleteD, self.channel_range, maxADC)

        # Create time data
        self.t = np.linspace(0, (self.totalSamples) * self.actualSampleIntervalNs, self.totalSamples)

        return self.t, self.adc2mVChAMax, self.adc2mVChBMax, self.adc2mVChCMax, self.adc2mVChDMax

    def save_value(self, dir):
        np.savez(dir, t=self.t, A=self.adc2mVChAMax, B=self.adc2mVChBMax, C=self.adc2mVChCMax, D=self.adc2mVChDMax)

    # code-related-digital Input
    def set_digital_port(self):
        # Set up digital port
        # handle = chandle
        # channel = PS3000A_DIGITAL_PORT0 = 0x80
        # enabled = 1
        # logicLevel = 10000
        self.status["SetDigitalPort"] = ps.ps3000aSetDigitalPort(self.chandle, ps.PS3000A_DIGITAL_PORT["PS3000A_DIGITAL_PORT0"], 1, 10000)
        # self.status["SetDigitalPort"] = ps.ps3000aSetDigitalPort(self.chandle, ps.PS3000A_DIGITAL_PORT["PS3000A_DIGITAL_PORT1"], 1, 10000)
        assert_pico_ok(self.status["SetDigitalPort"])

    def set_digital_buffer(self):
        # Create buffers ready for assigning pointers for data collection
        self.bufferDPort0Max = (ctypes.c_int16 * self.totalSamples)()
        self.bufferDPort0Min = (ctypes.c_int16 * self.totalSamples)()

        # Set the data buffer location for data collection from PS3000A_DIGITAL_PORT0
        # handle = chandle
        # source = PS3000A_DIGITAL_PORT0 = 0x80
        # Buffer max = ctypes.byref(bufferDPort0Max)
        # Buffer min = ctypes.byref(bufferDPort0Min)
        # Buffer length = totalSamples
        # Segment index = 0
        # Ratio mode = PS3000A_RATIO_MODE_NONE = 0
        self.status["SetDataBuffers"] = ps.ps3000aSetDataBuffers(self.chandle,
                                                                 ps.PS3000A_DIGITAL_PORT["PS3000A_DIGITAL_PORT0"],  # digital_port0,
                                                                 ctypes.byref(self.bufferDPort0Max),
                                                                 ctypes.byref(self.bufferDPort0Min),
                                                                 self.totalSamples,
                                                                 0,
                                                                 0)
        assert_pico_ok(self.status["SetDataBuffers"])

    def get_digital_values_simple(self):
        # Obtain binary for Digital Port 0
        # The tuple returned contains the channels in order (D7, D6, D5, ... D0).
        cTotalSamples = ctypes.c_int32(self.totalSamples)
        self.bufferDPort0 = splitMSODataFast(cTotalSamples, self.bufferDPort0Max)

        return self.bufferDPort0

    def Stream_AD(self):
        # Begin streaming mode:
        sampleInterval = ctypes.c_int32(self.interval)
        # print(sampleInterval.value)
        sampleUnits = ps.PS3000A_TIME_UNITS['PS3000A_NS']  # PS3000A_US, PS3000A_NS
        # We are not triggering:
        maxPreTriggerSamples = 0
        autoStopOn = 1
        # No downsampling:
        downsampleRatio = 1
        self.status["runStreaming"] = ps.ps3000aRunStreaming(
            self.chandle,
            ctypes.byref(sampleInterval),
            sampleUnits,
            maxPreTriggerSamples,
            self.totalSamples,
            autoStopOn,
            downsampleRatio,
            ps.PS3000A_RATIO_MODE['PS3000A_RATIO_MODE_NONE'],
            self.sizeOfOneBuffer
        )
        assert_pico_ok(self.status["runStreaming"])

        actualSampleInterval = sampleInterval.value
        # print(sampleInterval.value)
        self.actualSampleIntervalNs = actualSampleInterval * 1000

        print("Capturing at sample interval %s ns" % self.actualSampleIntervalNs)

        # We need a big buffer, not registered with the driver, to keep our complete capture in.
        self.bufferCompleteA = np.zeros(shape=self.totalSamples, dtype=np.int16)
        self.bufferCompleteB = np.zeros(shape=self.totalSamples, dtype=np.int16)
        self.bufferCompleteC = np.zeros(shape=self.totalSamples, dtype=np.int16)
        self.bufferCompleteD = np.zeros(shape=self.totalSamples, dtype=np.int16)
        self.bufferCompleteDport0M = np.zeros(shape=self.totalSamples, dtype=np.int16)
        self.bufferCompleteDport0m = np.zeros(shape=self.totalSamples, dtype=np.int16)

        self.nextSample = 0
        self.autoStopOuter = False
        self.wasCalledBack = False
        # Convert the python function into a C function pointer.
        cFuncPtr = ps.StreamingReadyType(self._streaming_callback_AD)

        # Fetch data from the driver in a loop,
        # copying it out of the registered buffers and into our complete one.

        while self.nextSample < self.totalSamples and not self.autoStopOuter:
            self.wasCalledBack = False
            self.status["getStreamingLastestValues"] = ps.ps3000aGetStreamingLatestValues(
                self.chandle,
                cFuncPtr,
                None
            )
            if not self.wasCalledBack:
                # If we weren't called back by the driver, this means no data is ready.
                # Sleep for a short while before trying
                # again.
                time.sleep(0.01)

        print("Done grabbing values.")
        print("Capturing interval %s sec" % (actualSampleInterval * 1000 * self.totalSamples / 1e12))

    def _streaming_callback_AD(self, handle, noOfSamples, startIndex, overflow, triggerAt, triggered, autoStop, param):
        # global nextSample, autoStopOuter, wasCalledBack
        self.wasCalledBack = True
        destEnd = self.nextSample + noOfSamples
        sourceEnd = startIndex + noOfSamples
        self.bufferCompleteA[self.nextSample:destEnd] = self.bufferAMax[startIndex:sourceEnd]
        self.bufferCompleteB[self.nextSample:destEnd] = self.bufferBMax[startIndex:sourceEnd]
        self.bufferCompleteC[self.nextSample:destEnd] = self.bufferCMax[startIndex:sourceEnd]
        self.bufferCompleteD[self.nextSample:destEnd] = self.bufferDMax[startIndex:sourceEnd]
        self.bufferCompleteDport0M[self.nextSample:destEnd] = self.bufferDPort0Max[startIndex:sourceEnd]
        self.bufferCompleteDport0m[self.nextSample:destEnd] = self.bufferDPort0Min[startIndex:sourceEnd]
        self.nextSample += noOfSamples
        if autoStop:
            self.autoStopOuter = True



import socket
import time


class PID:
    """Class to communicate with the Omega CNi16D54-EIT PID controller.

    Args:
        pid_ip (str): IP address for the controller.
        pid_port (int): Port number for the socket connection.
        verb (bool): Verbose output setting. Defaults to False.

    Attributes:
        verb (bool): Verbose output setting.
        hex_freq (str): Currently declared rotation frequency in hexadecimal.
        direction (int): Current direction of the HWP. 0 for forward and 1 for
            backwards.
        conn (socket.socket): Socket object with open connection to the PID
            controller.

    """

    def __init__(self, pid_ip, pid_port, verb=False):
        self.verb = verb
        self.hex_freq = '00000'
        self.direction = None
        # Need to setup connection before setting direction
        self.conn = self._establish_connection(pid_ip, int(pid_port))
        self.set_direction('0')

    @staticmethod
    def _establish_connection(ip, port, timeout=5):
        """Connect to PID controller.

        Args:
            ip (str): IP address for controller.
            port (int): Port number for socket connection.
            timeout (float): Time in seconds to wait for comms until timeout
                occurs.

        Returns:
            socket.socket: Socket object with open connection to the PID
                controller.

        """
        conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # unit tests might fail on first connection attempt
        attempts = 3
        for attempt in range(attempts):
            try:
                conn.connect((ip, port))
                break
            except ConnectionRefusedError:
                print(f"Failed to connect to device at {ip}:{port}")
                print(f"Connection attempts remaining: {attempts-attempt-1}")
            time.sleep(1)
        conn.settimeout(timeout)

        return conn

    @staticmethod
    def _convert_to_hex(value, decimal):
        """Converts the user input into a format the PID controller can
        read.

        """
        temp_value = hex(int(10**decimal * float(value)))
        return ('0000' + str(temp_value)[2:].upper())[-4:]

    @staticmethod
    def _get_scale_hex(num, corr):
        """Gets the exponent in scientific notation."""
        expo = int(str(num * 10**40 + 0.01).split('+')[1]) - 40
        digits = round(num * 10**(-expo + 4))

        expo_hex = str(hex(corr - expo))[2:]
        digits_hex = ('00000' + str(hex(digits))[2:])[-5:]

        return expo_hex + digits_hex

    ######################################################################
    # Main Processes
    ######################################################################

    def set_direction(self, direction):
        """Sets the direction if the CHWP.

        Modifies the ``direction`` attribute.

        Args:
            direction (int): 0 for forward and 1 for backwards

        """
        if direction == '0':
            if self.verb:
                print('Forward')
            resp = self.send_message("*W02400000")
            self.direction = 0
        elif direction == '1':
            if self.verb:
                print('Reverse')
            resp = self.send_message("*W02401388")
            self.direction = 1

        if self.verb:
            print(self.return_messages([resp])[0])

    def declare_freq(self, freq):
        """Declare to memory what the CHWP frequency should be.

        **Note:** Does not actually change the frequency.

        Args:
            freq (float): Set freqency in Hz. Must be less than 3.5.

        """
        if float(freq) <= 3.5:
            self.hex_freq = '0' + self._convert_to_hex(freq, 3)
            if self.verb:
                print('Frequency Setpoint = ' + str(freq) + ' Hz')
        else:
            if self.verb:
                print('Invalid Frequency')

    def tune_stop(self):
        """Set the setpoint to 0 Hz and stop the CHWP."""
        if self.verb:
            print('Starting Stop')

        responses = []
        responses.append(self.send_message("*W0C83"))
        responses.append(self.send_message("*W01400000"))
        responses.append(self.send_message("*R01"))
        responses.append(self.send_message("*Z02"))
        if self.verb:
            print(responses)
            print(self.return_messages(responses))

        stop_params = [0.2, 0, 0]
        self.set_pid(stop_params)

    def tune_freq(self):
        """Set the setpoint to currently declared frequency.

        To declare the frequency, use ``PID.declare_freq()``.

        """
        if self.verb:
            print('Starting Tune')

        responses = []
        responses.append(self.send_message("*W0C81"))
        responses.append(self.send_message(f"*W014{self.hex_freq}"))
        responses.append(self.send_message("*R01"))
        responses.append(self.send_message("*Z02"))
        if self.verb:
            print(responses)
            print(self.return_messages(responses))

        tune_params = [0.2, 63, 0]
        self.set_pid(tune_params)

    def get_freq(self):
        """Returns the current frequency of the CHWP."""
        if self.verb:
            print('Finding CHWP Frequency')

        responses = []
        responses.append(self.send_message("*X01"))
        if self.verb:
            print(responses)

        freq = self.return_messages(responses)[0]
        return freq

    def get_direction(self):
        """Get the current rotation direction.

        Returns:
            int: 0 for forward and 1 for backwards

        """
        if self.verb:
            print('Finding CHWP Direction')

        responses = []
        responses.append(self.send_message("*R02"))

        direction = self.return_messages(responses)[0]
        if direction == 1:
            print('Direction = Reverse')
        elif direction == 0:
            print('Direction = Forward')
        self.direction = direction

        return direction

    def set_pid(self, params):
        """Sets the PID parameters of the controller."""
        if self.verb:
            print('Setting PID Params')

        p_value = self._convert_to_hex(params[0], 3)
        i_value = self._convert_to_hex(params[1], 0)
        d_value = self._convert_to_hex(params[2], 1)

        responses = []
        responses.append(self.send_message(f"*W17{p_value}"))
        responses.append(self.send_message(f"*W18{i_value}"))
        responses.append(self.send_message(f"*W19{d_value}"))
        responses.append(self.send_message("*Z02"))
        if self.verb:
            print(responses)
            print(self.return_messages(responses))

    def set_scale(self, slope, offset):
        """Set the conversion between feedback voltage and approximate
        frequency.

        """
        slope_hex = self._get_scale_hex(slope, 1)
        offset_hex = self._get_scale_hex(offset, 2)

        responses = []
        responses.append(self.send_message(f"*W14{slope_hex}"))
        responses.append(self.send_message(f"*W03{offset_hex}"))
        responses.append(self.send_message("*Z02"))
        if self.verb:
            print(responses)
            print(self.return_messages(responses))

    ######################################################################
    # Messaging
    ######################################################################

    def send_message(self, msg):
        """Send message over TCP to the PID controller.

        Args:
            msg (str): Command to send to the controller.

        Returns:
            str: Respnose from the controller.

        """
        self.conn.sendall((msg + '\r\n').encode())
        time.sleep(0.5)  # Don't send messages too quickly
        for attempt in range(2):
            try:
                data = self.conn.recv(4096).decode().strip()
                break
            except socket.timeout:
                print("Caught timeout waiting for response from PID controller. "
                      + "Trying again...")
                time.sleep(1)
                if attempt == 1:
                    raise RuntimeError(
                        'Response from PID controller timed out.')
        return data

    def return_messages(self, msg):
        """Decode list of responses from PID controller and return useful
        values.

        Args:
            msg (list): List of messages to decode.

        Returns:
            list: Decoded responses.

        """
        return self._decode_array(msg)

    @staticmethod
    def _decode_array(input_array):
        output_array = list(input_array)

        for index, string in enumerate(list(input_array)):
            header = string[0]

            if header == 'R':
                output_array[index] = PID._decode_read(string)
            elif header == 'W':
                output_array[index] = PID._decode_write(string)
            elif header == 'E':
                output_array[index] = 'PID Enabled'
            elif header == 'D':
                output_array[index] = 'PID Disabled'
            elif header == 'P':
                pass
            elif header == 'G':
                pass
            elif header == 'X':
                output_array[index] = PID._decode_measure(string)
            else:
                pass

        return output_array

    @staticmethod
    def _decode_read(string):
        read_type = string[1:3]
        if read_type == '01':
            return 'Setpoint = ' + str(int(string[4:], 16) / 1000.)
        # Decode direction
        if read_type == '02':
            if int(string[4:], 16) / 1000. > 2.5:
                print('Direction = Reverse')
                return 1
            else:
                print('Direction = Forward')
                return 0
        else:
            return 'Unrecognized Read'

    @staticmethod
    def _decode_write(string):
        write_type = string[1:]
        if write_type == '01':
            return 'Changed Setpoint'
        if write_type == '02':
            return 'Changed Direction'
        if write_type == '0C':
            return 'Changed Action Type'
        else:
            return 'Unrecognized Write'

    @staticmethod
    def _decode_measure(string):
        measure_type = string[1:3]
        if measure_type == '01':
            return float(string[3:])
        else:
            return 9.999



import time

import serial

operational_status_key = [
    'No probe',
    'Field overload',
    'New field reading',
    'Alarm',
    'Invalid probe',
    'None',
    'Calibration error',
    'Zero probe done'
]


class LakeShore425:
    def __init__(self, COM):
        self.port = serial.Serial(
            COM,
            baudrate=57600,
            bytesize=serial.SEVENBITS,
            parity=serial.PARITY_ODD,
            stopbits=serial.STOPBITS_ONE,
            timeout=5,
            xonxoff=False,
            rtscts=False,
            write_timeout=None,
            dsrdtr=False,
            inter_byte_timeout=None
        )

    def close(self):
        self.port.close()

    def command(self, command):
        tmp = command + "\r\n"
        self.port.write(tmp.encode('utf-8'))
        time.sleep(.1)

    def query(self, command):
        self.command(command)
        return self.port.readline().strip().decode('utf-8')

    # Commands/Queries
    def get_id(self):
        return self.query("*IDN?")

    def get_op_status(self):
        val = int(self.query("OPST?"))
        out = 'Operational Status: '
        for i in range(8):
            if (val >> i & 1):
                out += operational_status_key[i]
                out += ', '
        return out

    def get_field(self):
        return float(self.query("RDGFIELD?"))

    def set_zero(self):
        self.command("ZCLEAR")
        self.command("ZPROBE")



import argparse
import os
import random
import threading
import time
from contextlib import contextmanager

import numpy as np
import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

from socs.Lakeshore.Lakeshore370 import LS370


class YieldingLock:
    """A lock protected by a lock.  This braided arrangement guarantees
    that a thread waiting on the lock will get priority over a thread
    that has just released the lock and wants to reacquire it.

    The typical use case is a Process that wants to hold the lock as
    much as possible, but occasionally release the lock (without
    sleeping for long) so another thread can access a resource.  The
    method release_and_acquire() is provided to make this a one-liner.

    """

    def __init__(self, default_timeout=None):
        self.job = None
        self._next = threading.Lock()
        self._active = threading.Lock()
        self._default_timeout = default_timeout

    def acquire(self, timeout=None, job=None):
        if timeout is None:
            timeout = self._default_timeout
        if timeout is None or timeout == 0.:
            kw = {'blocking': False}
        else:
            kw = {'blocking': True, 'timeout': timeout}
        result = False
        if self._next.acquire(**kw):
            if self._active.acquire(**kw):
                self.job = job
                result = True
            self._next.release()
        return result

    def release(self):
        self.job = None
        return self._active.release()

    def release_and_acquire(self, timeout=None):
        job = self.job
        self.release()
        return self.acquire(timeout=timeout, job=job)

    @contextmanager
    def acquire_timeout(self, timeout=None, job='unnamed'):
        result = self.acquire(timeout=timeout, job=job)
        if result:
            try:
                yield result
            finally:
                self.release()
        else:
            yield result


class LS370_Agent:
    """Agent to connect to a single Lakeshore 370 device.

    Args:
        name (ApplicationSession): ApplicationSession for the Agent.
        port (str): Serial port for the 370 device, e.g. '/dev/ttyUSB2'
        fake_data (bool, optional): generates random numbers without connecting
            to LS if True.
        dwell_time_delay (int, optional): Amount of time, in seconds, to
            delay data collection after switching channels. Note this time
            should not include the change pause time, which is automatically
            accounted for. Will automatically be reduced to dwell_time - 1
            second if it is set longer than a channel's dwell time. This
            ensures at least one second of data collection at the end of a scan.

    """

    def __init__(self, agent, name, port, fake_data=False, dwell_time_delay=0):

        # self._acq_proc_lock is held for the duration of the acq Process.
        # Tasks that require acq to not be running, at all, should use
        # this lock.
        self._acq_proc_lock = TimeoutLock()

        # self._lock is held by the acq Process only when accessing
        # the hardware but released occasionally so that (short) Tasks
        # may run.  Use a YieldingLock to guarantee that a waiting
        # Task gets activated preferentially, even if the acq thread
        # immediately tries to reacquire.
        self._lock = YieldingLock(default_timeout=5)

        self.name = name
        self.port = port
        self.fake_data = fake_data
        self.dwell_time_delay = dwell_time_delay
        self.module = None
        self.thermometers = []

        self.log = agent.log
        self.initialized = False
        self.take_data = False

        self.agent = agent
        # Registers temperature feeds
        agg_params = {
            'frame_length': 10 * 60  # [sec]
        }
        self.agent.register_feed('temperatures',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    @ocs_agent.param('force', default=False, type=bool)
    def init_lakeshore(self, session, params=None):
        """init_lakeshore(auto_acquire=False, force=False)

        **Task** - Perform first time setup of the Lakeshore 370 communication.

        Parameters:
            auto_acquire (bool, optional): Default is False. Starts data
                acquisition after initialization if True.
            force (bool, optional): Force re-initialize the lakeshore if True.

        """

        if self.initialized and not params['force']:
            self.log.info("Lakeshore already initialized. Returning...")
            return True, "Already initialized"

        with self._lock.acquire_timeout(job='init') as acquired1, \
                self._acq_proc_lock.acquire_timeout(timeout=0., job='init') \
                as acquired2:
            if not acquired1:
                self.log.warn(f"Could not start init because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"
            if not acquired2:
                self.log.warn(f"Could not start init because "
                              f"{self._acq_proc_lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            if self.fake_data:
                self.res = random.randrange(1, 1000)
                session.add_message("No initialization since faking data")
                self.thermometers = ["thermA", "thermB"]
            else:
                self.module = LS370(self.port)
                print("Initialized Lakeshore module: {!s}".format(self.module))
                session.add_message("Lakeshore initilized with ID: %s" % self.module.id)

                self.thermometers = [channel.name for channel in self.module.channels]

            self.initialized = True

        # Start data acquisition if requested
        if params['auto_acquire']:
            self.agent.start('acq')

        return True, 'Lakeshore module initialized.'

    @ocs_agent.param('_')
    def acq(self, session, params=None):
        """acq()

        **Process** - Run data acquisition.

        """

        with self._acq_proc_lock.acquire_timeout(timeout=0, job='acq') \
                as acq_acquired, \
                self._lock.acquire_timeout(job='acq') as acquired:
            if not acq_acquired:
                self.log.warn(f"Could not start Process because "
                              f"{self._acq_proc_lock.job} is already running")
                return False, "Could not acquire lock"
            if not acquired:
                self.log.warn(f"Could not start Process because "
                              f"{self._lock.job} is holding the lock")
                return False, "Could not acquire lock"

            session.set_status('running')
            self.log.info("Starting data acquisition for {}".format(self.agent.agent_address))
            previous_channel = None
            last_release = time.time()

            self.take_data = True
            while self.take_data:

                # Relinquish sampling lock occasionally.
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self._lock.release_and_acquire(timeout=10):
                        self.log.warn(f"Failed to re-acquire sampling lock, "
                                      f"currently held by {self._lock.job}.")
                        continue

                if self.fake_data:
                    data = {
                        'timestamp': time.time(),
                        'block_name': 'fake-data',
                        'data': {}
                    }
                    for therm in self.thermometers:
                        reading = np.random.normal(self.res, 20)
                        data['data'][therm] = reading
                    time.sleep(.1)

                else:
                    active_channel = self.module.get_active_channel()

                    # The 370 reports the last updated measurement repeatedly
                    # during the "pause change time", this results in several
                    # stale datapoints being recorded. To get around this we
                    # query the pause time and skip data collection during it
                    # if the channel has changed (as it would if autoscan is
                    # enabled.)
                    if previous_channel != active_channel:
                        if previous_channel is not None:
                            pause_time = active_channel.get_pause()
                            self.log.debug("Pause time for {c}: {p}",
                                           c=active_channel.channel_num,
                                           p=pause_time)

                            dwell_time = active_channel.get_dwell()
                            self.log.debug("User set dwell_time_delay: {p}",
                                           p=self.dwell_time_delay)

                            # Check user set dwell time isn't too long
                            if self.dwell_time_delay > dwell_time:
                                self.log.warn("WARNING: User set dwell_time_delay of "
                                              + "{delay} s is larger than channel "
                                              + "dwell time of {chan_time} s. If "
                                              + "you are autoscanning this will "
                                              + "cause no data to be collected. "
                                              + "Reducing dwell time delay to {s} s.",
                                              delay=self.dwell_time_delay,
                                              chan_time=dwell_time,
                                              s=dwell_time - 1)
                                total_time = pause_time + dwell_time - 1
                            else:
                                total_time = pause_time + self.dwell_time_delay

                            for i in range(total_time):
                                self.log.debug("Sleeping for {t} more seconds...",
                                               t=total_time - i)
                                time.sleep(1)

                        # Track the last channel we measured
                        previous_channel = self.module.get_active_channel()

                    # Setup feed dictionary
                    channel_str = active_channel.name.replace(' ', '_')
                    data = {
                        'timestamp': time.time(),
                        'block_name': channel_str,
                        'data': {}
                    }

                    # Collect both temperature and resistance values from each Channel
                    data['data'][channel_str + '_T'] = \
                        self.module.get_temp(unit='kelvin', chan=active_channel.channel_num)
                    data['data'][channel_str + '_R'] = \
                        self.module.get_temp(unit='ohms', chan=active_channel.channel_num)

                    # Courtesy in case active channel has not changed
                    time.sleep(0.1)

                session.app.publish_to_feed('temperatures', data)

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        Stops acq process.
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'

    @ocs_agent.param('heater', type=str)
    @ocs_agent.param('range')
    @ocs_agent.param('wait', type=float, default=0)
    def set_heater_range(self, session, params):
        """set_heater_range(range, heater='sample', wait=0)

        **Task** - Adjust the heater range for servoing cryostat. Wait for a
        specified amount of time after the change.

        Parameters:
            heater (str): Name of heater to set range for, 'sample' by default
                (and the only implemented option.)
            range (str, float): see arguments in
                :func:`socs.Lakeshore.Lakeshore370.Heater.set_heater_range`
            wait (float, optional): Amount of time to wait after setting the
                heater range. This allows the servo time to adjust to the new range.

        """
        with self._lock.acquire_timeout(job='set_heater_range') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            heater_string = params.get('heater', 'sample')
            if heater_string.lower() == 'sample':
                heater = self.module.sample_heater
            elif heater_string.lower() == 'still':  # TODO: add still heater class to driver
                # heater = self.module.still_heater
                self.log.warn(f"{heater_string} heater not yet implemented in this agent, please modify client")

            current_range = heater.get_heater_range()

            if params['range'] == current_range:
                print("Current heater range matches commanded value. Proceeding unchanged.")
            else:
                heater.set_heater_range(params['range'])
                time.sleep(params['wait'])

        return True, f'Set {heater_string} heater range to {params["range"]}'

    @ocs_agent.param('channel', type=int, check=lambda x: 1 <= x <= 16)
    @ocs_agent.param('mode', type=str, choices=['current', 'voltage'])
    def set_excitation_mode(self, session, params):
        """set_excitation_mode(channel, mode)

        **Task** - Set the excitation mode of a specified channel.

        Parameters:
            channel (int): Channel to set the excitation mode for. Valid values
                are 1-16.
            mode (str): Excitation mode. Possible modes are 'current' or
                'voltage'.

        """
        with self._lock.acquire_timeout(job='set_excitation_mode') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            self.module.chan_num2channel(params['channel']).set_excitation_mode(params['mode'])
            session.add_message(f'post message in agent for Set channel {params["channel"]} excitation mode to {params["mode"]}')
            print(f'print statement in agent for Set channel {params["channel"]} excitation mode to {params["mode"]}')

        return True, f'return text for Set channel {params["channel"]} excitation mode to {params["mode"]}'

    @ocs_agent.param('channel', type=int, check=lambda x: 1 <= x <= 16)
    @ocs_agent.param('value', type=float)
    def set_excitation(self, session, params):
        """set_excitation(channel, value)

        **Task** - Set the excitation voltage/current value of a specified
        channel.

        Parameters:
            channel (int): Channel to set the excitation for. Valid values
                are 1-16.
            value (float): Excitation value in volts or amps depending on set
                excitation mode. See
                :func:`socs.Lakeshore.Lakeshore370.Channel.set_excitation`

        """
        with self._lock.acquire_timeout(job='set_excitation') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            current_excitation = self.module.chan_num2channel(params['channel']).get_excitation()

            if params['value'] == current_excitation:
                print(f'Channel {params["channel"]} excitation already set to {params["value"]}')
            else:
                self.module.chan_num2channel(params['channel']).set_excitation(params['value'])
                session.add_message(f'Set channel {params["channel"]} excitation to {params["value"]}')
                print(f'Set channel {params["channel"]} excitation to {params["value"]}')

        return True, f'Set channel {params["channel"]} excitation to {params["value"]}'

    @ocs_agent.param('P', type=int)
    @ocs_agent.param('I', type=int)
    @ocs_agent.param('D', type=int)
    def set_pid(self, session, params):
        """set_pid(P, I, D)

        **Task** - Set the PID parameters for servo control of fridge.

        Parameters:
            P (int): Proportional term for PID loop
            I (int): Integral term for the PID loop
            D (int): Derivative term for the PID loop

        Notes:
            Makes a call to :func:`socs.Lakeshore.Lakeshore370.Heater.set_pid`.

        """
        with self._lock.acquire_timeout(job='set_pid') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            self.module.sample_heater.set_pid(params["P"], params["I"], params["D"])
            session.add_message(f'post message text for Set PID to {params["P"]}, {params["I"]}, {params["D"]}')
            print(f'print text for Set PID to {params["P"]}, {params["I"]}, {params["D"]}')

        return True, f'return text for Set PID to {params["P"]}, {params["I"]}, {params["D"]}'

    @ocs_agent.param('channel', type=int)
    def set_active_channel(self, session, params):
        """set_active_channel(channel)

        **Task** - Set the active channel on the LS370.

        Parameters:
            channel (int): Channel to switch readout to. Valid values are 1-16.

        """
        with self._lock.acquire_timeout(job='set_active_channel') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            self.module.set_active_channel(params["channel"])
            session.add_message(f'post message text for set channel to {params["channel"]}')
            print(f'print text for set channel to {params["channel"]}')

        return True, f'return text for set channel to {params["channel"]}'

    @ocs_agent.param('autoscan', type=bool)
    def set_autoscan(self, session, params):
        """set_autoscan(autoscan)

        **Task** - Sets autoscan on the LS370.

        Parameters:
            autoscan (bool): True to enable autoscan, False to disable.

        """
        with self._lock.acquire_timeout(job='set_autoscan') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            if params['autoscan']:
                self.module.enable_autoscan()
                self.log.info('enabled autoscan')
            else:
                self.module.disable_autoscan()
                self.log.info('disabled autoscan')

        return True, 'Set autoscan to {}'.format(params['autoscan'])

    @ocs_agent.param('temperature', type=float, check=lambda x: x < 1)
    @ocs_agent.param('channel', type=float, default=None)
    def servo_to_temperature(self, session, params):
        """servo_to_temperature(temperature, channel=None)

        **Task** - Servo to a given temperature using a closed loop PID on a
        fixed channel. This will automatically disable autoscan if enabled.

        Parameters:
            temperature (float): Temperature to servo to in units of Kelvin.
            channel (int, optional): Channel to servo off of.

        """
        with self._lock.acquire_timeout(job='servo_to_temperature') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            # Check we're in correct control mode for servo.
            if self.module.sample_heater.mode != 'Closed Loop':
                session.add_message('Changing control to Closed Loop mode for servo.')
                self.module.sample_heater.set_mode("Closed Loop")

            # Check we aren't autoscanning.
            if self.module.get_autoscan() is True:
                session.add_message('Autoscan is enabled, disabling for PID control on dedicated channel.')
                self.module.disable_autoscan()

            # Check to see if we passed an input channel, and if so change to it
            if params.get("channel", None) is not None:
                session.add_message(f'Changing heater input channel to {params.get("channel")}')
                self.module.sample_heater.set_input_channel(params.get("channel"))

            # Check we're scanning same channel expected by heater for control.
            if self.module.get_active_channel().channel_num != int(self.module.sample_heater.input):
                session.add_message('Changing active channel to expected heater control input')
                self.module.set_active_channel(int(self.module.sample_heater.input))

            # Check we're setup to take correct units.
            if self.module.sample_heater.units != 'kelvin':
                session.add_message('Setting preferred units to Kelvin on heater control.')
                self.module.sample_heater.set_units('kelvin')

            # Make sure we aren't servoing too high in temperature.
            if params["temperature"] > 1:
                return False, 'Servo temperature is set above 1K. Aborting.'

            self.module.sample_heater.set_setpoint(params["temperature"])

        return True, f'Setpoint now set to {params["temperature"]} K'

    @ocs_agent.param('measurements', type=int)
    @ocs_agent.param('threshold', type=float)
    def check_temperature_stability(self, session, params):
        """check_temperature_stability(measurements, threshold)

        Check servo temperature stability is within threshold.

        Parameters:
            measurements (int): number of measurements to average for stability
                check
            threshold (float): amount within which the average needs to be to
                the setpoint for stability

        """
        with self._lock.acquire_timeout(job='check_temp_stability') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            setpoint = float(self.module.sample_heater.get_setpoint())

            if params is None:
                params = {'measurements': 10, 'threshold': 0.5e-3}

            test_temps = []

            for i in range(params['measurements']):
                test_temps.append(self.module.get_temp())
                time.sleep(.1)  # sampling rate is 10 readings/sec, so wait 0.1 s for a new reading

            mean = np.mean(test_temps)
            session.add_message(f'Average of {params["measurements"]} measurements is {mean} K.')
            print(f'Average of {params["measurements"]} measurements is {mean} K.')

            if np.abs(mean - setpoint) < params['threshold']:
                print("passed threshold")
                session.add_message('Setpoint Difference: ' + str(mean - setpoint))
                session.add_message(f'Average is within {params["threshold"]} K threshold. Proceeding with calibration.')

                return True, f"Servo temperature is stable within {params['threshold']} K"

            else:
                print("we're in the else")
                # adjust_heater(t,rest)

        return False, f"Temperature not stable within {params['threshold']}."

    @ocs_agent.param('heater', type=str, choices=['sample', 'still'])
    @ocs_agent.param('mode', type=str, choices=['Off', 'Monitor Out', 'Open Loop', 'Zone', 'Still', 'Closed Loop', 'Warm up'])
    def set_output_mode(self, session, params=None):
        """set_output_mode(heater, mode)

        **Task** - Set output mode of the heater.

        Parameters:
            heater (str): Name of heater to set range for, either 'sample' or
                'still'.
            mode (str): Specifies mode of heater. Can be "Off", "Monitor Out",
                "Open Loop", "Zone", "Still", "Closed Loop", or "Warm up"

        """
        with self._lock.acquire_timeout(job='set_output_mode') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            session.set_status('running')

            if params['heater'].lower() == 'still':
                # self.module.still_heater.set_mode(params['mode']) #TODO: add still heater to driver
                self.log.warn(f"{params['heater']} heater not yet implemented in this agent, please modify client")
            if params['heater'].lower() == 'sample':
                self.module.sample_heater.set_mode(params['mode'])
            self.log.info("Set {} output mode to {}".format(params['heater'], params['mode']))

        return True, "Set {} output mode to {}".format(params['heater'], params['mode'])

    @ocs_agent.param('heater', type=str, choices=['sample', 'still'])
    @ocs_agent.param('output', type=float)
    @ocs_agent.param('display', type=str, choices=['current', 'power'], default=None)
    def set_heater_output(self, session, params=None):
        """set_heater_output(heater, output, display=None)

        **Task** - Set display type and output of the heater.

        Parameters:
            heater (str): Name of heater to set range for, either 'sample' or
                'still'.
            output (float): Specifies heater output value. For possible values see
                :func:`socs.Lakeshore.Lakeshore370.Heater.set_heater_output`
            display (str, optional): Specifies heater display type. Can be
                "current" or "power". If None, heater display is not reset
                before setting output.

        """

        with self._lock.acquire_timeout(job='set_heater_output') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self._lock.job} is already running")
                return False, "Could not acquire lock"

            heater = params['heater'].lower()
            output = params['output']

            display = params.get('display', None)

            if heater == 'still':  # TODO: add still heater to driver
                # self.module.still_heater.set_heater_output(output, display_type=display)
                self.log.warn(f"{heater} heater not yet implemented in this agent, please modify client")
            if heater.lower() == 'sample':
                self.log.info("display: {}\toutput: {}".format(display, output))
                self.module.sample_heater.set_heater_output(output, display_type=display)

            self.log.info("Set {} heater display to {}, output to {}".format(heater, display, output))

            session.set_status('running')

            data = {'timestamp': time.time(),
                    'block_name': '{}_heater_out'.format(heater),
                    'data': {'{}_heater_out'.format(heater): output}
                    }
            session.app.publish_to_feed('temperatures', data)

        return True, "Set {} display to {}, output to {}".format(heater, display, output)

    @ocs_agent.param('attribute', type=str)
    @ocs_agent.param('channel', type=int, default=1)
    def get_channel_attribute(self, session, params):
        """get_channel_attribute(attribute, channel=1)

        **Task** - Gets an arbitrary channel attribute, stored in the session.data dict.

        Parameters:
            attribute (str, optional): Attribute to get from the 370.
            channel (int, optional): Channel to get the attribute for.

        Notes:
            Channel attributes stored in the session.data object are in the
            structure::

                >>> response.session['data']
                {"calibration_curve": 21,
                 "dwell": 3,
                 "excitation": 6.32e-6,
                 "excitation_mode": "voltage",
                 "excitation_power": 2.0e-15,
                 "kelvin_reading": 100.0e-3,
                 "pause": 3,
                 "reading_status": ["T.UNDER"]
                 "resistance_range": 2.0e-3,
                 "resistance_reading": 10.0e3,
                 "temperature_coefficient": "negative",
                }

            Only attribute called with this method will be populated for the
            given channel. This example shows all available attributes.

        """
        with self._lock.acquire_timeout(job=f"get_{params['attribute']}", timeout=3) as acquired:
            if not acquired:
                print(f"Lock could not be acquired because it is held by {self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get channel
            channel_key = int(params.get('channel', 1))
            channel = self.module.chan_num2channel(channel_key)

            # check that attribute is a valid channel method
            if getattr(channel, f"get_{params['attribute']}", False) is not False:
                query = getattr(channel, f"get_{params['attribute']}")

            # get attribute
            resp = query()
            session.data[params['attribute']] = resp

            time.sleep(.1)

        return True, f"Retrieved {channel.name} {params['attribute']}"

    @ocs_agent.param('attribute', type=str)
    def get_heater_attribute(self, session, params):
        """get_heater_attribute(attribute)

        **Task** - Gets an arbitrary heater attribute, stored in the session.data dict.

        Parameters:
            attribute (str): Heater attribute to get.

        Notes:
            Heater attributes stored in the session.data object are in the structure::

                >>> response.session['data']
                {"heater_range": 1e-3,
                 "heater_setup": ["current", 1e-3, 120],
                 "input_channel": 6,
                 "manual_out": 0.0,
                 "mode": "Closed Loop",
                 "pid": (80, 10, 0),
                 "setpoint": 100e-3,
                 "still_output", 10.607,
                 "units": "kelvin",
                }

            Only the attribute called with this method will be populated,
            this example just shows all available attributes.

        """
        with self._lock.acquire_timeout(job=f"get_{params['attribute']}", timeout=3) as acquired:
            if not acquired:
                print(f"Lock could not be acquired because it is held by {self._lock.job}")
                return False, 'Could not acquire lock'

            session.set_status('running')

            # get heater
            heater = self.module.sample_heater

            # check that attribute is a valid heater method
            if getattr(heater, f"get_{params['attribute']}", False) is not False:
                query = getattr(heater, f"get_{params['attribute']}")

            # get attribute
            resp = query()
            session.data[params['attribute']] = resp

            time.sleep(.1)

        return True, f"Retrieved sample heater {params['attribute']}"


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--port', type=str, help='Full path to USB node for the lakeshore, e.g. "/dev/ttyUSB0"')
    pgroup.add_argument('--serial-number')
    pgroup.add_argument('--mode')
    pgroup.add_argument('--fake-data', type=int, default=0,
                        help='Set non-zero to fake data, without hardware.')
    pgroup.add_argument('--dwell-time-delay', type=int, default=0,
                        help="Amount of time, in seconds, to delay data\
                              collection after switching channels. Note this\
                              time should not include the change pause time,\
                              which is automatically accounted for.\
                              Will automatically be reduced to dwell_time - 1\
                              second if it is set longer than a channel's dwell\
                              time. This ensures at least one second of data\
                              collection at the end of a scan.")
    pgroup.add_argument('--auto-acquire', type=bool, default=True,
                        help='Automatically start data acquisition on startup')

    return parser


def main(args=None):
    # For logging
    txaio.use_twisted()
    LOG = txaio.make_logger()

    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='Lakeshore370Agent',
                                  parser=parser,
                                  args=args)

    # Automatically acquire data if requested (default)
    init_params = False
    if args.auto_acquire:
        init_params = {'auto_acquire': True}

    LOG.info('I am in charge of device with serial number: %s' % args.serial_number)

    agent, runner = ocs_agent.init_site_agent(args)

    lake_agent = LS370_Agent(agent, args.serial_number, args.port,
                             fake_data=args.fake_data,
                             dwell_time_delay=args.dwell_time_delay)

    agent.register_task('init_lakeshore', lake_agent.init_lakeshore,
                        startup=init_params)
    agent.register_task('set_heater_range', lake_agent.set_heater_range)
    agent.register_task('set_excitation_mode', lake_agent.set_excitation_mode)
    agent.register_task('set_excitation', lake_agent.set_excitation)
    agent.register_task('set_pid', lake_agent.set_pid)
    agent.register_task('set_autoscan', lake_agent.set_autoscan)
    agent.register_task('set_active_channel', lake_agent.set_active_channel)
    agent.register_task('servo_to_temperature', lake_agent.servo_to_temperature)
    agent.register_task('check_temperature_stability', lake_agent.check_temperature_stability)
    agent.register_task('set_output_mode', lake_agent.set_output_mode)
    agent.register_task('set_heater_output', lake_agent.set_heater_output)
    agent.register_task('get_channel_attribute', lake_agent.get_channel_attribute)
    agent.register_task('get_heater_attribute', lake_agent.get_heater_attribute)
    agent.register_process('acq', lake_agent.acq, lake_agent._stop_acq)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



# Tucker Elleflot
from socs.common.prologix_interface import PrologixInterface


class PsuInterface(PrologixInterface):
    def __init__(self, ip_address, gpibAddr, verbose=False, **kwargs):
        self.verbose = verbose
        super().__init__(ip_address, gpibAddr, **kwargs)

    def enable(self, ch):
        '''
        Enables output for channel (1,2,3) but does not turn it on.
        Depending on state of power supply, it might need to be called
        before the output is set.
        '''
        self.set_chan(ch)
        self.write('OUTP:ENAB ON')

    def disable(self, ch):
        '''
        disabled output from a channel (1,2,3). once called, enable must be
        called to turn on the channel again
        '''
        self.write('OUTP:ENAB OFF')

    def set_chan(self, ch):
        self.write('inst:nsel ' + str(ch))

    def set_output(self, ch, out):
        '''
        set status of power supply channel
        ch - channel (1,2,3) to set status
        out - ON: True|1|'ON' OFF: False|0|'OFF'

        Calls enable to ensure a channel can be turned on. We might want to
        make them separate (and let us use disable as a safety feature) but
        for now I am thinking we just want to thing to turn on when we tell
        it to turn on.
        '''
        self.set_chan(ch)
        self.enable(ch)
        if isinstance(out, str):
            self.write('CHAN:OUTP ' + out)
        elif out:
            self.write('CHAN:OUTP ON')
        else:
            self.write('CHAN:OUTP OFF')

    def get_output(self, ch):
        '''
        check if the output of a channel (1,2,3) is on (True) or off (False)
        '''
        self.set_chan(ch)
        self.write('CHAN:OUTP:STAT?')
        out = bool(float(self.read()))
        return out

    def set_volt(self, ch, volt):
        self.set_chan(ch)
        self.write('volt ' + str(volt))
        if self.verbose:
            voltage = self.get_volt(ch)
            print("CH " + str(ch) + " is set to " + str(voltage) + " V")

    def set_curr(self, ch, curr):
        self.set_chan(ch)
        self.write('curr ' + str(curr))
        if self.verbose:
            current = self.get_curr(ch)
            print("CH " + str(ch) + " is set to " + str(current) + " A")

    def get_volt(self, ch):
        self.set_chan(ch)
        self.write('MEAS:VOLT? CH' + str(ch))
        voltage = float(self.read())
        return voltage

    def get_curr(self, ch):
        self.set_chan(ch)
        self.write('MEAS:CURR? CH' + str(ch))
        current = float(self.read())
        return current



import os
import time
from datetime import datetime, timezone


def openlog(log_dir):
    now = datetime.now(timezone.utc)
    dateStr = now.strftime('%Y-%m-%d')

    logfilename = '{}/OCS_{}.dat'.format(log_dir, dateStr)
    if not os.path.isdir(log_dir):
        os.mkdir(log_dir)
    if os.path.exists(logfilename):
        logfile = open(logfilename, 'a+')
    else:
        logfile = open(logfilename, 'w')
        log = '#UnixTime ON/OFF time-period[sec] position[def] mode\n'
        logfile.write(log)
        logfile.flush()
    return logfile


# onoff: 'ON' or 'OFF'
# position: encoder current angle [deg.]
# mode: 'calibration', 'stepwise', 'continuous'
def writelog(logfile, onoff, timeperiod=0., position=0., mode='stepping_rot'):
    if timeperiod > 0.:
        log = '{:.6f} {:3s} {:8.3f} {:8.3f} {:14s}\n'\
              .format(time.time(), onoff, timeperiod, position, mode)
    else:
        log = '{:.6f} {:3s} {:8s} {:8.3f} {:14s}\n'\
              .format(time.time(), onoff, '--------', position, mode)
    logfile.write(log)
    logfile.flush()
    return True



import argparse
import datetime
import glob
import os
import re
import threading
import time

import txaio
from ocs import ocs_agent, site_config

# Notes:
# Each Lakeshore log gets its own "block" due to tracking the logs
# being somewhat tricky. If a block structure is established and
# then the thermometer removed from the scan, say if the user
# switches to scanning a single channel, then the block structure
# won't be reconstructed and thus won't match the existing
# structure, causing an error.

# For logging
txaio.use_twisted()
LOG = txaio.make_logger()


class LogTracker:
    """Log Tracking helper class. Always tracks current date's logs.

    Parameters
    ----------
    log_dir : str
        Top level log directory

    Attributes
    ----------
    log_dir : str
        Top level log directory
    date : datetime.date
        Today's date. Used to determine the active log directory
    file_objects : dict
        A dictionary with filenames as keys, and another dict as the value.
        Each of these sub-dictionaries has two keys, "file_object", and
        "stat_results", with the open file object, and os.stat results as
        values, respectively. For example::

            {'CH6 T 21-05-27.log':
                {'file_object': <_io.TextIOWrapper name='CH6 T 21-05-27.log' mode='r' encoding='UTF-8'>,
                'stat_results': os.stat_result(st_mode=33188,
                                               st_ino=1456748,
                                               st_dev=65024,
                                               st_nlink=1,
                                               st_uid=1000,
                                               st_gid=1000,
                                               st_size=11013,
                                               st_atime=1622135813,
                                               st_mtime=1622135813,
                                               st_ctime=1622135813)
                }
            }

    """

    def __init__(self, log_dir):
        self.log_dir = log_dir
        self.date = datetime.date.fromtimestamp(time.time())
        self.file_objects = {}

    def _build_file_list(self):
        """Get list of files to open.

        Assumes all logs are set to log to same top level directory,
        i.e. /home/bluefors/logs/

        """
        date_str = self.date.strftime("%y-%m-%d")
        file_list = glob.glob("{}/{}/*.log".format(self.log_dir,
                                                   date_str))
        return file_list

    def _open_file(self, filename):
        """Open a single file, adding it to self.file_objects.

        Parameters
        ----------
        filename : str
            Full path to filename to open

        """
        if filename not in self.file_objects.keys():
            print("{} not yet open, opening...".format(filename))
            self.file_objects[filename] = {"file_object": open(filename, 'r'),
                                           "stat_results": os.stat(filename)}
        else:
            pass

    def check_open_files(self):
        """Check all log files are opened.

        The Status logs don't always exist, but we want to catch them when they
        do get generated. This rebuilds the file list and checks all the files in it
        are in the open file objects dictionary.
        """
        file_list = self._build_file_list()
        for f in file_list:
            self._open_file(f)

    def set_active_date(self):
        """Set the active date to today."""
        new_date = datetime.date.fromtimestamp(time.time())

        if new_date > self.date:
            self.close_all_files()
            self.date = new_date
            self.open_all_logs()

    def reopen_file(self, filename):
        """If a file is already open and a new inode is detected, reopen the
        file. Returns the last line of the file.

        Parameters
        ----------
        filename : str
            Full path to filename to open

        """
        self.file_objects[filename] = {"file_object": open(filename, 'r'),
                                       "stat_results": os.stat(filename)}
        lines = self.file_objects[filename]["file_object"].readlines()
        return lines[-1]

    def open_all_logs(self):
        """Open today's logs and move to end of files."""
        file_list = self._build_file_list()

        self.file_objects = {}
        for _file in file_list:
            self._open_file(_file)

        for k, v in self.file_objects.items():
            v['file_object'].readlines()

    def close_all_files(self):
        """Close all the files tracked by the LogTracker."""
        for k, v in self.file_objects.items():
            v['file_object'].close()
            print("Closed file: {}".format(k))

        self.file_objects = {}


class LogParser:
    """Log Parsing helper class.

    Knows the internal formats for each log type. Used to loop over all
    logs tracked by a LogTracker and publish their contents to an OCS Feed.

    Parameters
    ----------
    tracker : LogTracker
        log tracker that contains paths and file objects to parse
    mode : str
        Operating mode for the log tracker. Either "follow" or "poll",
        defaulting to "follow". In "follow" mode the Tracker will read the
        next line in the file if able to. In "poll" mode stats about the
        file are used to determine if it was updated since the last read,
        and if it has been the file is reopened to get the last line. This
        is more I/O intensive, but is useful in certain configurations.
    stale_time : int
        Time in minutes which represents how fresh data in the bluefors
        logs must be when we open them in order to publish to OCS. This
        ensures we don't reopen a file much later than when they were
        collected and publish "stale" data to the OCS live HK system.

    """

    def __init__(self, tracker, mode="follow", stale_time=2):
        self.log_tracker = tracker
        self.patterns = {'channels': ['v11', 'v2', 'v1', 'turbo1', 'v12', 'v3', 'v10',
                                      'v14', 'v4', 'v13', 'compressor', 'v15', 'v5',
                                      'hs-still', 'v21', 'v16', 'v6', 'scroll1', 'v17',
                                      'v7', 'scroll2', 'v18', 'v8', 'pulsetube', 'v19',
                                      'v20', 'v9', 'hs-mc', 'ext'],
                         'status': ['tc400errorcode', 'tc400ovtempelec',
                                    'tc400ovtemppump', 'tc400setspdatt',
                                    'tc400pumpaccel', 'tc400commerr',
                                    'tc400errorcode_2', 'tc400ovtempelec_2',
                                    'tc400ovtemppump_2', 'tc400setspdatt_2',
                                    'tc400pumpaccel_2', 'tc400commerr_2',
                                    'nxdsf', 'nxdsct', 'nxdst', 'nxdsbs', 'nxdstrs',
                                    'ctrl_pres', 'cpastate', 'cparun', 'cpawarn',
                                    'cpaerr', 'cpatempwi', 'cpatempwo', 'cpatempo',
                                    'cpatemph', 'cpalp', 'cpalpa', 'cpahp', 'cpahpa',
                                    'cpadp', 'cpacurrent', 'cpahours', 'cpapscale',
                                    'cpascale', 'cpatscale', 'cpasn', 'cpamodel',
                                    'ctrl_pres_ok', 'ctr_pressure_ok'],
                         'heater': ["a1_u", "a1_r_lead", "a1_r_htr", "a2_u",
                                    "a2_r_lead", "a2_r_htr", "htr", "htr_range"]}
        self.mode = mode
        self.stale_time = stale_time

    @staticmethod
    def timestamp_from_str(time_string):
        """Convert time string from Bluefors log file into a UNIX timestamp.

        Parameters
        ----------
        time_string : str
            String in format "%d-%m-%y,%H:%M:%S"

        Returns
        -------
        float
            UNIX timestamp

        """
        dt = datetime.datetime.strptime(time_string, "%d-%m-%y,%H:%M:%S")
        timestamp = dt.timestamp()
        return timestamp

    def _parse_single_value_log(self, new_line, log_name):
        """Parses a simple single value log. Valid for LS372 and flowmeter log
        files.

        Parameters
        ----------
        new_line : str
            The new line read by the Parser
        log_name : str
            The name of the log, returned by self.identify_log()
        """
        date, _time, data_value = new_line.strip().split(',')
        time_str = "%s,%s" % (date, _time)
        timestamp = self.timestamp_from_str(time_str)

        # Data array to populate
        data = {
            'timestamp': timestamp,
            'block_name': log_name,
            'data': {}
        }

        data['data'][log_name] = float(data_value)

        return data

    def _parse_maxigauge_log(self, new_line, log_name):
        """Parse the maxigauge logs.

        Parameters
        ----------
        new_line : str
            The new line read by the Parser
        log_name : str
            The name of the log, returned by self.identify_log()

        """
        ts, *channels = new_line.strip().split('CH')
        time_str = ts.strip(',')
        timestamp = self.timestamp_from_str(time_str)

        ch_data = {}
        for ch in channels:
            ch_num, _, state, value, *_ = ch.split(',')
            ch_data["pressure_ch{}_state".format(ch_num)] = int(state)
            ch_data["pressure_ch{}".format(ch_num)] = float(value)

        data = {
            'timestamp': timestamp,
            'block_name': log_name,
            'data': ch_data
        }

        return data

    def _parse_multi_value_log(self, new_line, log_type, log_name):
        """Parse a log containing multiple values on each line. Valid for
        Channel, Status, and heater logs.

        Patterns to search for are all known by the parser a class attribute.

        Parameters
        ----------
        new_line : str
            The new line read by the Parser
        log_type : str
            Log type as identified by self.identify_log, used to select search
            patterns
        log_name : str
            The name of the log, returned by self.identify_log()

        """
        date, _time, *_ = new_line.strip().split(',')
        time_str = "%s,%s" % (date, _time)
        timestamp = self.timestamp_from_str(time_str)

        data_array = {}
        for pattern in self.patterns[log_type]:
            regex = re.compile(rf'{pattern},([0-9\.\+\-E]+)')
            m = regex.search(new_line)

            # skip patterns that don't exist in this log
            if not m:
                continue

            if log_type == 'channels':
                data_array[pattern.replace('-', '_')] = int(m.group(1))
            else:
                data_array[pattern] = float(m.group(1))

        data = {
            'timestamp': timestamp,
            'block_name': log_name,
            'data': data_array
        }

        return data

    @staticmethod
    def identify_log(filename):
        """Identify type of log return unique identifier

        Parameters
        ----------
        filename : str
            path to file for identification

        Returns
        -------
        tuple
            A tuple containing two strings:
                1. The log type, i.e. 'lakeshore', 'flowmeter'
                2. Unique identifier based on filetype, i.e. 'lakeshore_ch8_p',
                   'pressure_ch1_state'

        """
        # file type: regex search pattern to identify
        file_types = {'lakeshore': '(CH[0-9]+ [T,R,P])',
                      'flowmeter': '(Flowmeter)',
                      'maxiguage': '(maxigauge)',
                      'channels': '(Channels)',
                      'status': '(Status)',
                      'errors': '(Errors)',
                      'heater': '(heaters)'}

        for k, v in file_types.items():
            if re.search(v, filename):
                _type = k

                m = re.search(file_types[_type], filename)
                if _type == 'lakeshore':
                    return _type, "{}_{}".format(_type, m.group(0).lower().replace(' ', '_'))
                else:
                    return _type, m.group(0).lower()

        # If nothing matches return None
        return (None, None)

    def read_and_publish_logs(self, app_session):
        """Read a new line from each log file if there is one, and publish its
        contents to the app_session's feed.

        Parameters
        ----------
        app_session : ocs.ocs_agent.OpSession
            session from the ocs_agent, used to publish to bluefors feed

        """
        for k, v in self.log_tracker.file_objects.items():
            log_type, log_name = self.identify_log(k)

            if os.stat(k).st_ino != v['stat_results'].st_ino and self.mode == "poll":
                LOG.debug("New inode found, reopening...")
                new = self.log_tracker.reopen_file(k)
                LOG.debug("File: {f}, Line: {l}", f=k, l=new)
            # In a situation with a samba share mounted via sshfs, reading the
            # nextline didn't reliably work, nor does watching the inode. We'll
            # also check modification times, which maybe we should just do
            # instead of the inode check...
            elif os.stat(k).st_mtime > v['stat_results'].st_mtime and self.mode == "poll":
                LOG.debug("Modification detected, reopening...")
                new = self.log_tracker.reopen_file(k)
                LOG.debug("File: {f}, Line: {l}", f=k, l=new)
            else:
                new = v['file_object'].readline()
            if new == '':
                continue

            if log_type in ['lakeshore', 'flowmeter']:
                data = self._parse_single_value_log(new, log_name)
            elif log_type == 'maxiguage':
                data = self._parse_maxigauge_log(new, log_name)
            elif log_type in ['channels', 'status', 'heater']:
                data = self._parse_multi_value_log(new, log_type, log_name)
            elif log_type == 'errors':
                LOG.info("The Bluefors Agent cannot process error logs. "
                         + "Not publishing.")
                data = None
            else:
                LOG.warn("Warning: Unknown log type. Skipping publish step. "
                         + "This probably shouldn't happen.")
                LOG.warn("Filename: {}".format(k))
                data = None

            if data is not None:
                LOG.debug("Data: {d}", d=data)
                # If the file was reopened due to an inode change we don't know
                # if the last line is recent enough to be worth publishing. Check
                if (time.time() - data['timestamp']) < int(self.stale_time) * 60:
                    app_session.app.publish_to_feed('bluefors', data)
                else:
                    LOG.warn("Not publishing stale data. Make sure your log "
                             + "file sync is done at a rate faster than once ever "
                             + "{x} minutes.", x=self.stale_time)


class BlueforsAgent:
    """Agent to track the Bluefors logs generated by an LD400.

    Parameters
    ----------
    log_directory : str
        Top level log directory for the Bluefors logs

    """

    def __init__(self, agent, log_directory):
        self.lock = threading.Semaphore()
        self.job = None

        self.log_tracker = LogTracker(log_directory)

        self.log = agent.log
        self.agent = agent

        # Registers bluefors feed
        agg_params = {
            'frame_length': float(os.environ.get("FRAME_LENGTH", 10 * 60))  # [sec]
        }
        self.log.debug("frame_length set to {length}",
                       length=agg_params['frame_length'])
        self.agent.register_feed('bluefors',
                                 record=True,
                                 agg_params=agg_params,
                                 )

    def try_set_job(self, job_name):
        print(self.job, job_name)
        with self.lock:
            if self.job is None:
                self.job = job_name
                return (True, 'ok')
            else:
                return (False, 'Conflict: "%s" is already running.' % self.job)

    def set_job_done(self):
        with self.lock:
            self.job = None

    def acq(self, session, params=None):
        """acq()

        **Process** - Monitor and publish data from the Bluefors log files.

        """

        ok, msg = self.try_set_job('acq')
        if not ok:
            return ok, msg

        session.set_status('running')

        # Create file objects for all logs in today's directory
        self.log_tracker.open_all_logs()

        # Determine parser configuration
        stale_time = os.environ.get("STALE_TIME", 2)
        mode = os.environ.get("MODE", "follow")

        # Setup the Parser object with tracking info
        parser = LogParser(self.log_tracker, mode, stale_time)

        while True:
            with self.lock:
                if self.job == '!acq':
                    break
                elif self.job == 'acq':
                    pass
                else:
                    return 10

            # Make sure we're looking at today's logs
            self.log_tracker.set_active_date()

            # Ensure all the logs we want are open
            self.log_tracker.check_open_files()

            # Check for new lines and publish to feed
            parser.read_and_publish_logs(session)

            time.sleep(0.01)

        self.set_job_done()
        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        ok = False
        with self.lock:
            if self.job == 'acq':
                self.job = '!acq'
                ok = True
        return (ok, {True: 'Requested process stop.',
                     False: 'Failed to request process stop.'}[ok])


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--log-directory')

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    # Setup argument parser
    parser = make_parser()
    args = site_config.parse_args(agent_class='BlueforsAgent',
                                  parser=parser,
                                  args=args)
    LOG.info('I am following logs located at : %s' % args.log_directory)

    agent, runner = ocs_agent.init_site_agent(args)

    bluefors_agent = BlueforsAgent(agent, args.log_directory)

    agent.register_process('acq', bluefors_agent.acq,
                           bluefors_agent._stop_acq, startup=True)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import os
import time
from collections import deque
from enum import Enum
from os import environ

import numpy as np
import txaio

# For logging
txaio.use_twisted()

ON_RTD = os.environ.get('READTHEDOCS') == 'True'
if not ON_RTD:
    import so3g  # noqa: F401
    from ocs import ocs_agent, site_config
    from spt3g import core


class FlowControl(Enum):
    """Flow control enumeration."""
    ALIVE = 0
    START = 1
    END = 2
    CLEANSE = 3


# Could extend FlowControl, but this keeps this simulator only thing separate
SHUTDOWN = 'shutdown'


class StreamChannel:
    """Simulated SMuRF channel for stream testing.

    Uses np.random.normal to generate random Gaussian data.

    Parameters
    ----------
    mean : float
        Mean value of Gaussian to simulate data with
    stdev : float
        Standard deviation of Gaussian to simulate data

    """

    def __init__(self, mean, stdev):
        self.mean = mean
        self.stdev = stdev

    def read(self):
        """Read a value from the channel.

        Returns
        -------
        float
            Random, normally distributed, value

        """
        return np.random.normal(self.mean, self.stdev)


class SmurfStreamSimulator:
    """OCS Agent to simulate data streaming without connection to a SMuRF.

    Parameters
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    target_host : str
        Target remote host address
    port : int
        Port to send data over
    num_chans : int
        Number of channels to simulate
    stream_id : str
        Stream ID to put into G3Frames. Defaults to "stream_sim"

    Attributes
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    log : txaio.tx.Logger
        txaio logger ojbect, created by the OCSAgent
    target_host : str
        Target remote host address
    port : int
        Port to send data over
    writer : spt3g.core.G3NetworkSender
        G3 writer for sending frames to, in this case a G3NetworkSender
    is_streaming : bool
        flag to track if we're currently streaming, used in stop task to stop
        the streaming process. If stopped then keep alive flow control frames
        are still being sent.
    running_in_background : bool
        flag to track if the streaming process is running in the background,
    channels : list
        List of simulated channels to stream

    """

    def __init__(self, agent, target_host="*", port=4536, num_chans=528,
                 stream_id='stream_sim'):
        self.agent = agent
        self.log = agent.log
        self.target_host = target_host

        self.port = port

        self.stream_id = stream_id

        self.writer = None
        self.is_streaming = False
        self.running_in_background = False

        self.channels = [
            StreamChannel(0, 1) for i in range(num_chans)
        ]

    def start_background_streamer(self, session, params=None):
        """start_background_streamer(params=None)

        Process to run streaming process. A data stream is started
        automatically. It can be stopped and started by the start and stop
        tasks. Either way keep alive flow control frames are being sent.

        Parameters
        ----------
        frame_rate : float, optional
            Frequency [Hz] at which G3Frames are sent over the network.
            Defaults to 1 frame pers sec.
        sample_rate : float, optional
            Sample rate [Hz] for each channel. Defaults to 10 Hz.

        """
        if params is None:
            params = {}

        self.writer = core.G3NetworkSender(hostname=self.target_host,
                                           port=self.port)

        frame_rate = params.get('frame_rate', 1.)
        sample_rate = params.get('sample_rate', 10.)

        frame_num = 0
        self.running_in_background = True

        # Control flags FIFO stack to keep Writer single threaded
        self.flags = deque([FlowControl.START])

        while self.running_in_background:
            # Send START frame
            if next(iter(self.flags), None) is FlowControl.START:
                self._set_stream_on()  # sends start flowcontrol
                self.is_streaming = True
                self.flags.popleft()

            print("stream running in background")
            self.log.debug("control flags: {f}", f=self.flags)
            # Send keep alive flow control frame
            f = core.G3Frame(core.G3FrameType.none)
            f['sostream_flowcontrol'] = FlowControl.ALIVE.value
            self.writer.Process(f)

            if self.is_streaming:
                frame_start = time.time()
                time.sleep(1. / frame_rate)
                frame_stop = time.time()
                times = np.arange(frame_start, frame_stop, 1. / sample_rate)

                f = core.G3Frame(core.G3FrameType.Scan)
                f['session_id'] = 0
                f['frame_num'] = frame_num
                f['sostream_id'] = self.stream_id
                f['data'] = core.G3TimestreamMap()

                for i, chan in enumerate(self.channels):
                    ts = core.G3Timestream([chan.read() for t in times])
                    ts.start = core.G3Time(frame_start * core.G3Units.sec)
                    ts.stop = core.G3Time(frame_stop * core.G3Units.sec)
                    f['data'][f"r{i:04}"] = ts

                self.writer.Process(f)
                self.log.info("Writing frame...")
                frame_num += 1

                # Send END frame
                if next(iter(self.flags), None) is FlowControl.END:
                    self._send_end_flowcontrol_frame()
                    self._send_cleanse_flowcontrol_frame()
                    self.is_streaming = False
                    self.flags.popleft()

            else:
                # Don't send keep alive frames too quickly
                time.sleep(1)

            # Shutdown streamer
            if next(iter(self.flags), None) is SHUTDOWN:
                self.running_in_background = False
                self.flags.popleft()

        # Teardown writer
        self.writer.Close()
        self.writer = None

        return True, "Finished streaming"

    # Not thread safe, be aware of where you're calling these
    def _send_start_flowcontrol_frame(self):
        """Send START flowcontrol frame."""
        if self.writer is not None:
            self.log.info("Sending START flowcontrol frame")
            f = core.G3Frame(core.G3FrameType.none)
            f['sostream_flowcontrol'] = FlowControl.START.value
            self.writer.Process(f)

            # Send new Observation frame at start of observation
            f = core.G3Frame(core.G3FrameType.Observation)
            f['session_id'] = 0
            f['start_time'] = time.time()
            f['sostream_id'] = self.stream_id
            self.writer.Process(f)

    def _send_end_flowcontrol_frame(self):
        """Send END flowcontrol frame."""
        if self.writer is not None:
            self.log.info("Sending END flowcontrol frame")
            f = core.G3Frame(core.G3FrameType.none)
            f['sostream_flowcontrol'] = FlowControl.END.value
            self.writer.Process(f)

    def _send_cleanse_flowcontrol_frame(self):
        """Send CLEANSE flowcontrol frames."""
        if self.writer is not None:
            self.log.info("Sending CLEANSE flowcontrol frame")

            f = core.G3Frame(core.G3FrameType.Observation)
            f['sostream_flowcontrol'] = FlowControl.CLEANSE.value
            self.writer.Process(f)

            f = core.G3Frame(core.G3FrameType.Wiring)
            f['sostream_flowcontrol'] = FlowControl.CLEANSE.value
            self.writer.Process(f)

    def _set_stream_on(self):
        """Private method that that isn't a "task".

        Starts the stream of actual data frames from the background streaming
        process.

        """
        self._send_start_flowcontrol_frame()
        self.is_streaming = True

    # Thread safe
    def stop_background_streamer(self, session, params=None):
        """stop_background_streamer(params=None)

        Stop method associated with start_background_streamer process.

        The design here sets a flag that is checked in the main
        background_streamer process, that way we keep the G3Writer writing from
        a single thread.

        """
        if self.is_streaming:
            if FlowControl.END not in self.flags:
                self.flags.append(FlowControl.END)

        if SHUTDOWN not in self.flags:
            self.flags.append(SHUTDOWN)
        return True, "Stopping stream"

    def set_stream_on(self, session, params=None):
        """set_stream_on(params=None)

        Task to start the stream of actual data frames from the background
        streaming process.

        The design here sets a flag that is checked in the main
        background_streamer process, that way we keep the G3Writer writing from
        a single thread.

        Parameters
        ----------
        force : bool
            Whether to force a start frame or not, defaults to False

        """
        if params is None:
            params = {}

        force = params.get('force', False)
        if force:
            self.log.debug("force starting stream")

        if not self.is_streaming or force:
            if FlowControl.START not in self.flags:
                self.flags.append(FlowControl.START)

        return True, "Started stream"

    def set_stream_off(self, session, params=None):
        """set_stream_off(params=None)

        Task to stop the stream of actual data frames from the background
        streaming process. Keep alive flow control frames will still be sent.

        The design here sets a flag that is checked in the main
        background_streamer process, that way we keep the G3Writer writing from
        a single thread.

        Parameters
        ----------
        force : bool
            Whether to force a start frame or not, defaults to False

        """
        if params is None:
            params = {}

        force = params.get('force', False)

        if self.is_streaming or force:
            if FlowControl.END not in self.flags:
                self.flags.append(FlowControl.END)

        return True, "Stopped stream"


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group("Agent Options")
    pgroup.add_argument("--auto-start", default=False, type=bool,
                        help="Automatically start streaming at "
                        + "Agent startup.")
    pgroup.add_argument("--target-host", default="*",
                        help="Target remote host.")
    pgroup.add_argument("--port", default=50000,
                        help="Port to listen on.")
    pgroup.add_argument("--num-chans", default=528,
                        help="Number of detector channels to simulate.")
    pgroup.add_argument("--stream-id", default="stream_sim",
                        help="Stream ID for the simulator.")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='SmurfStreamSimulator',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    sim = SmurfStreamSimulator(agent, target_host=args.target_host,
                               port=int(args.port),
                               num_chans=int(args.num_chans),
                               stream_id=args.stream_id)

    agent.register_process('stream', sim.start_background_streamer,
                           sim.stop_background_streamer,
                           startup=bool(args.auto_start))
    agent.register_task('start', sim.set_stream_on)
    agent.register_task('stop', sim.set_stream_off)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import array as arr
import struct
import time

from serial import Serial

# some commands require a CRC code (cyclic redundancy check) -
# these require the provided CRC table
crc_table = arr.array('H', [
    0x0, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
    0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
    0x1231, 0x210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
    0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
    0x2462, 0x3443, 0x420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
    0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
    0x3653, 0x2672, 0x1611, 0x630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
    0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
    0x48c4, 0x58e5, 0x6886, 0x78a7, 0x840, 0x1861, 0x2802, 0x3823,
    0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
    0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0xa50, 0x3a33, 0x2a12,
    0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
    0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0xc60, 0x1c41,
    0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
    0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0xe70,
    0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
    0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
    0x1080, 0xa1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
    0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
    0x2b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
    0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
    0x34e2, 0x24c3, 0x14a0, 0x481, 0x7466, 0x6447, 0x5424, 0x4405,
    0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
    0x26d3, 0x36f2, 0x691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
    0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
    0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x8e1, 0x3882, 0x28a3,
    0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
    0x4a75, 0x5a54, 0x6a37, 0x7a16, 0xaf1, 0x1ad0, 0x2ab3, 0x3a92,
    0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
    0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0xcc1,
    0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
    0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0xed1, 0x1ef0
])


def calc_crc(data):
    """Calculates CRC."""

    crc = 0
    for i in data:
        crc = crc_table[(crc >> 8) ^ i] ^ (crc << 8)
        crc = crc % 65536
    return crc


class VantagePro2:
    """Allows communication to Vantage Pro 2 Weather Monitor Module.
    Contains commands to be issued and member variables that store
    collected data.
    """

    def __init__(self, path, baud=19200, timeout=1):
        """Establish serial connection and initialize member variables."""

        if not path:
            path = '/dev/ttyUSB0'
        self.com = Serial(port=path, baudrate=baud, timeout=timeout)
        self.startup()

    def startup(self):
        """Wakeup vantage pro 2 console."""
        self.com.write(b"\n")
        for i in range(0, 3):
            response = self.com.read(2)
            if response == b'\n\r':
                break
            time.sleep(1.2)
        else:
            raise TimeoutError("Vantage Pro 2 console not woke")

    def crc_check(self, data):
        """Checks received data. If received data has a CRC value of 0, then
        correct message received!

        """

        crc = calc_crc(data)
        if crc != 0:
            print('Failed CRC. Errors in data received')

    def close(self):
        self.com.close()

    # closing serial if agent goes out of scope

    def __exit__(self):
        self.com.close()

    def conditions_screen(self):
        """Move console to conditions screen (where loop command can
        succesfully be sent.)

        """

        self.startup()
        self.com.write(b'RXTEST\n')
        time.sleep(2)

        # reads response from console
        for i in range(0, 3):
            ok = self.com.read(6)
            if ok == b'\n\rOK\n\r':
                break
        else:
            raise TimeoutError("Command not acknowledged")

    def msg(self, msg):
        """Send general command or query to module. If command received and
        acknowledged, continue else retry 2 more times.

        """

        attempt = 0
        while attempt < 3:
            self.startup()
            self.com.write(bytes(msg, 'ascii'))
            time.sleep(0.1)
            ack = 0
            for i in range(0, 3):
                ack = self.com.read(1)
                if ack == b'\x06':
                    return True
            attempt += 1
        else:
            return False

    def interrupt_daq(self):
        """Interrupts loop command...if sent before loop command finishes."""

        self.startup()
        self.com.write(b'\n')
        for i in range(0, 3):
            response = self.com.read(2)
            if response == b'\n\r':
                print("Data Acquisition succesfully interrupted")
                break
            time.sleep(0.1)
        else:
            raise TimeoutError('Data Acquisition not interrupted')

    def receive_data(self):
        """Reads weather data from console and returns loop_data{}.
        User should read Vantage Pro Serial Communication Reference Manual
        for the format of the loop data packet.
        Specifically for units of data! Available fields:

        * Barometer trend: the current 3 hour barometer trend. Possible values::

                -60 = Falling rapidly
                -20 = Falling slowly
                  0 = Steady
                 20 = Rising slowly
                 60 = Rising rapidly
                 80 = No trend available

                Any other value = The VantagePro2 does not have the 3 hours
                of data needed to determine the barometer trend.

        * Barometer: Current barometer reading (Hg/1000)
        * Inside Temperature: Temperatue in Fahrenheit (10th of a degree)
        * Inside Humidity: Relative humidity in percent
        * Outside Temperature: Temperature in Fahrenheit (10th of a degree)
        * Wind Speed: Wind speed in miles per hour
        * 10 min average wind speed: 10 minute average wind speed in mph
        * Wind Direction: From 1-360 degrees. Possible values::

                  0 = No wind direction data
                 90 = East
                180 = South
                270 = West
                360 = North

        * Extra Temperatures: Temperature from up to 7 extra temperature
          stations.
        * Soil Temperatures: Four soil temperature sensors, in the same
          format as the extra temperatures format listed above.
        * Leaf Temperatures: Four leaf temperature sensors, in the same
          format as the extra temperatures format listed above.
        * Outside Humidity: Relativie humidity in percent
        * Extra Humidities: Realtive humidity in percent for 7 humidity
          stations
        * Rain Rate: Number of rain clicks (0.1in or 0.2mm) per hour
        * UV: "Unit is in UV index"
        * Solar Radiation: Units in watt/meter^2
        * Storm Rain: Stored in 100th of an inch
        * Start Date of Current storm: Gives month, date,
          and year (offset by 2000)
        * Day Rain: Number of rain clicks (0.1in or 0.2mm)/hour
          in the past day
        * Month Rain: Number of rain clicks (0.1in or 0.2mm)/hour
          in the past month
        * Year Rain: Number of rain clicks (0.1in or 0.2mm)/hour
          in the past year
        * Day ET: 1000th of an inch
        * Month ET: 1000th of an inch
        * Year ET: 1000th of an inch
        * Soil Moistures: In centibar, supports 4 soil sensors
        * Leaf Wetnesses: Scale from 0-15. Supports 4 leaf sensors. Possible
          values::

                 0 = Very dry
                15 = Very wet

        * Inside Alarms: Currently active inside alarms
        * Rain Alarms: Currently active rain alarms
        * Outside Alarms: Currently active outside alarms
        * Outside Humidity Alarms: Currently active humidity alarms
        * Extra Temp/Hum Alarms: Currently active extra
          temperature/humidity alarms
        * Soil & Leaf Alarms: Currently active soil/leaf alarms
        * Console Battery Voltage: Voltage
        * Time of Sunrise: Time is stored as hour x 100 + min
        * Time of Sunset: Time is stored as hour x 100 + min

        """

        info = b''

        # Give device multiple chances to send its data
        for i in range(0, 3):
            info = self.com.read(99)
            if info:
                break
            else:
                time.sleep(0.5)
        else:
            raise TimeoutError("Timeout error: no data from LOOP command")

        loop_data = {}
        byte_data = struct.unpack('=5b3h1b1h2B1H23b1h1b9h25b1h2b2h2c1h', info)
        loop_data['bar_trend'] = byte_data[3]
        loop_data['packet_type'] = byte_data[4]
        loop_data['next_record'] = byte_data[5]
        loop_data['barometer'] = byte_data[6] / 1000.0
        loop_data['temp_inside'] = byte_data[7] / 10.0
        loop_data['humidity_inside'] = byte_data[8]
        loop_data['temp_outside'] = byte_data[9] / 10.0
        loop_data['wind_speed'] = byte_data[10]
        loop_data['avg_wind_speed'] = byte_data[11]
        loop_data['wind_dir'] = byte_data[12]
        loop_data['extra_temp0'] = byte_data[13] - 90.0
        loop_data['extra_temp1'] = byte_data[14] - 90.0
        loop_data['extra_temp2'] = byte_data[15] - 90.0
        loop_data['extra_temp3'] = byte_data[16] - 90.0
        loop_data['extra_temp4'] = byte_data[17] - 90.0
        loop_data['extra_temp5'] = byte_data[18] - 90.0
        loop_data['extra_temp6'] = byte_data[19] - 90.0
        loop_data['soil_temp0'] = byte_data[20] - 90.0
        loop_data['soil_temp1'] = byte_data[21] - 90.0
        loop_data['soil_temp2'] = byte_data[22] - 90.0
        loop_data['soil_temp3'] = byte_data[23] - 90.0
        loop_data['leaf_temp0'] = byte_data[24] - 90.0
        loop_data['leaf_temp1'] = byte_data[25] - 90.0
        loop_data['leaf_temp2'] = byte_data[26] - 90.0
        loop_data['leaf_temp3'] = byte_data[27] - 90.0
        loop_data['humidity_outside'] = byte_data[28]
        loop_data['extra_hum0'] = byte_data[29]
        loop_data['extra_hum1'] = byte_data[30]
        loop_data['extra_hum2'] = byte_data[31]
        loop_data['extra_hum3'] = byte_data[32]
        loop_data['extra_hum4'] = byte_data[33]
        loop_data['extra_hum5'] = byte_data[34]
        loop_data['extra_hum6'] = byte_data[35]
        loop_data['rain_rate'] = byte_data[36]
        loop_data['UV'] = byte_data[37]
        loop_data['solar_rad'] = byte_data[38]
        loop_data['storm_rain'] = byte_data[39] / 100.0
        loop_data['storm_start'] = byte_data[40]
        loop_data['day_rain'] = byte_data[41]
        loop_data['month_rain'] = byte_data[42]
        loop_data['year_rain'] = byte_data[43]
        loop_data['day_ET'] = byte_data[44] / 1000.0
        loop_data['month_ET'] = byte_data[45] / 100.0
        loop_data['year_ET'] = byte_data[46] / 100.0
        loop_data['soil_moisture0'] = byte_data[47]
        loop_data['soil_moisture1'] = byte_data[48]
        loop_data['soil_moisture2'] = byte_data[49]
        loop_data['soil_moisture3'] = byte_data[50]
        loop_data['leaf_wetness0'] = byte_data[51]
        loop_data['leaf_wetness1'] = byte_data[52]
        loop_data['leaf_wetness2'] = byte_data[53]
        loop_data['leaf_wetness3'] = byte_data[54]
        loop_data['inside_alarm0'] = byte_data[55]
        loop_data['inside_alarm1'] = byte_data[56]
        loop_data['rain_alarm'] = byte_data[57]
        loop_data['outside_alarm0'] = byte_data[58]
        loop_data['outside_alarm1'] = byte_data[59]
        loop_data['extra_temp_hum_alarm0'] = byte_data[60]
        loop_data['extra_temp_hum_alarm1'] = byte_data[61]
        loop_data['extra_temp_hum_alarm2'] = byte_data[62]
        loop_data['extra_temp_hum_alarm3'] = byte_data[63]
        loop_data['extra_temp_hum_alarm4'] = byte_data[64]
        loop_data['extra_temp_hum_alarm5'] = byte_data[65]
        loop_data['extra_temp_hum_alarm6'] = byte_data[66]
        loop_data['soil_leaf_alarm0'] = byte_data[67]
        loop_data['soil_leaf_alarm1'] = byte_data[68]
        loop_data['soil_leaf_alarm2'] = byte_data[69]
        loop_data['soil_leaf_alarm3'] = byte_data[70]
        loop_data['transmitter_battery_status'] = byte_data[71]
        loop_data['console_battery_voltage'] = ((byte_data[72] * 300) / 512) / 100.0
        loop_data['forecast_icons'] = byte_data[73]
        loop_data['forecast_rule_num'] = byte_data[74]
        loop_data['time_sunrise'] = byte_data[75]
        loop_data['time_sunset'] = byte_data[76]

        # CRC check, data must be sent byte by byte
        pure_data = struct.unpack('=99b', info)
        self.crc_check(pure_data)

        return loop_data

    def weather_daq(self):
        """Issues "LOOP 1"(which samples weather data) command to weather
        station, and unpacks the data.

        """

        # Startup and issue loop command
        self.startup()
        command = "LOOP 1\n"
        if not self.msg(command):
            print("message failed to be sent!")

        # collect data
        return self.receive_data()

    def print_data(self, data):
        """Prints contents of data collected from LOOP <loops> command.
        Loop: loop#
        Field : Value format.
        """

        print("**************")
        for i in data:
            print("{} : {}".format(i, data[i]))
